__author__ = 'Bohdan Mushkevych'

from datetime import datetime

from synergy.db.model import unit_of_work
from synergy.db.model.synergy_mq_transmission import SynergyMqTransmission
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao
from synergy.mq.flopsy import PublishersPool
from synergy.workers.abstract_mq_worker import AbstractMqWorker
from synergy.system.performance_tracker import UowAwareTracker
from synergy.scheduler.scheduler_constants import QUEUE_UOW_REPORT


class AbstractUowAwareWorker(AbstractMqWorker):
    """ Abstract class is inherited by all workers/aggregators
    that are aware of unit_of_work and capable of processing it"""

    def __init__(self, process_name):
        super(AbstractUowAwareWorker, self).__init__(process_name)
        self.uow_dao = UnitOfWorkDao(self.logger)
        self.publishers = PublishersPool(self.logger)

    def __del__(self):
        self.publishers.close()
        super(AbstractUowAwareWorker, self).__del__()

    # **************** Abstract Methods ************************
    def _init_performance_ticker(self, logger):
        self.performance_ticker = UowAwareTracker(logger)
        self.performance_ticker.start()

    def _process_uow(self, uow):
        """
        :param uow: unit_of_work to process
        :return: a tuple (number of processed items/documents/etc, desired unit_of_work state) or None
        if None is returned then it is assumed that the return tuple is (0, unit_of_work.STATE_PROCESSED)
        :raise an Exception if the UOW shall be marked as STATE_INVALID
        """
        raise NotImplementedError('method _process_uow must be implemented by {0}'.format(self.__class__.__name__))

    def _clean_up(self):
        """ method is called from the *finally* clause and is suppose to clean up after the uow processing """
        pass

    def _mq_callback(self, message):
        try:
            mq_request = SynergyMqTransmission.from_json(message.body)
            uow = self.uow_dao.get_one(mq_request.unit_of_work_id)
            if not uow.is_requested:
                # accept only UOW in STATE_REQUESTED
                self.logger.warn('Skipping UOW: id {0}; state {1};'.format(message.body, uow.state),
                                 exc_info=False)
                self.consumer.acknowledge(message.delivery_tag)
                return
        except Exception:
            self.logger.error('Safety fuse. Can not identify UOW {0}'.format(message.body), exc_info=True)
            self.consumer.acknowledge(message.delivery_tag)
            return

        try:
            uow.state = unit_of_work.STATE_IN_PROGRESS
            uow.started_at = datetime.utcnow()
            self.uow_dao.update(uow)
            self.performance_ticker.start_uow(uow)

            result = self._process_uow(uow)
            if result is None:
                self.logger.warn('method {0}._process_uow returned None. Assuming happy flow.'
                                 .format(self.__class__.__name__))
                number_of_aggregated_objects, target_state = 0, unit_of_work.STATE_PROCESSED
            else:
                number_of_aggregated_objects, target_state = result

            uow.number_of_aggregated_documents = number_of_aggregated_objects
            uow.number_of_processed_documents = self.performance_ticker.success_per_job
            uow.finished_at = datetime.utcnow()
            uow.state = target_state
            self.uow_dao.update(uow)

            if uow.is_finished:
                self.performance_ticker.finish_uow()
            else:
                self.performance_ticker.cancel_uow()

        except Exception as e:
            fresh_uow = self.uow_dao.get_one(mq_request.unit_of_work_id)
            self.performance_ticker.cancel_uow()
            if fresh_uow.is_canceled:
                self.logger.warn('UOW {0} for {1}@{2} was likely marked by MX as SKIPPED. No UOW update is performed.'
                                 .format(uow.db_id, uow.process_name, uow.timeperiod), exc_info=False)
            else:
                self.logger.error('Safety fuse while processing UOW {0} for {1}@{2}: {3}'
                                  .format(uow.db_id, uow.process_name, uow.timeperiod, e), exc_info=True)
                uow.state = unit_of_work.STATE_INVALID
                self.uow_dao.update(uow)

        finally:
            self.consumer.acknowledge(message.delivery_tag)
            self.consumer.close()
            self._clean_up()

        try:
            publisher = self.publishers.get(QUEUE_UOW_REPORT)
            publisher.publish(mq_request.document)
            publisher.release()
            self.logger.info('Published UOW status report into queue {0}'.format(QUEUE_UOW_REPORT))
        except Exception:
            self.logger.error('Error on UOW status report publishing', exc_info=True)



# Scheduler constants
LAG_5_MINUTES = 5  # lag in minutes between finish of the timeperiod and beginning of its processing

PROCESS_SCHEDULER = 'Scheduler'
PROCESS_GC = 'GarbageCollector'
PROCESS_MX = 'MX'

TOKEN_SCHEDULER = 'scheduler'
TOKEN_GC = 'gc'
TOKEN_WERKZEUG = 'werkzeug'

# State Machine names
STATE_MACHINE_RECOMPUTING = 'recomputing'
STATE_MACHINE_CONTINUOUS = 'continuous'
STATE_MACHINE_DISCRETE = 'discrete'
STATE_MACHINE_FREERUN = 'freerun'

# dependency allows processing of the dependent timeperiod,
# however finalization of the dependent timeperiod is not allowed until all blocking timeperiods are processed
BLOCKING_NORMAL = 'blocking_normal'

# any processing of dependent timeperiods is blocked until blocking timeperiods are processed.
BLOCKING_DEPENDENCIES = 'blocking_dependencies'

# any processing of higher time granularity is blocked until all children timeperiods are processed
BLOCKING_CHILDREN = 'blocking_children'

# MQ exchanges that are required by Synergy Scheduler
EXCHANGE_MANAGED_WORKER = 'ex_managed_worker'
EXCHANGE_FREERUN_WORKER = 'ex_freerun_worker'
EXCHANGE_UTILS = 'ex_utils'

# DB Collection/Table names
COLLECTION_MANAGED_PROCESS = 'managed_process'
COLLECTION_FREERUN_PROCESS = 'freerun_process'
COLLECTION_UNIT_OF_WORK = 'unit_of_work'

COLLECTION_JOB_HOURLY = 'job_hourly'
COLLECTION_JOB_DAILY = 'job_daily'
COLLECTION_JOB_MONTHLY = 'job_monthly'
COLLECTION_JOB_YEARLY = 'job_yearly'

QUEUE_UOW_REPORT = 'q_uow_report'



__author__ = 'Bohdan Mushkevych'

import setproctitle

from synergy.conf import settings
from synergy.system.data_logging import get_logger
from synergy.system.utils import create_pid_file, remove_pid_file


class SynergyProcess(object):
    """ Fundamental class for all processes. Registers logger and renames process to SynergyYYY"""

    def __init__(self, process_name):
        """ renames process to SynergyYYY and creates PID file """
        self.process_name = process_name
        self.logger = get_logger(process_name)

        # process-related activities
        setproctitle.setproctitle(settings.settings['process_prefix'] + self.process_name)
        create_pid_file(self.process_name)

    def __del__(self):
        """ removes PID file """
        remove_pid_file(self.process_name)
        self.logger.info('Shutdown {0}'.format(self.process_name))



__author__ = 'Bohdan Mushkevych'

from werkzeug.utils import cached_property
from synergy.mx.base_request_handler import BaseRequestHandler
from synergy.mx.rest_model_factory import create_rest_timetable_tree, create_rest_managed_scheduler_entry


class TreeDetails(BaseRequestHandler):
    def __init__(self, request, **values):
        super(TreeDetails, self).__init__(request, **values)

    def _get_tree_details(self, tree_name):
        tree_obj = self.scheduler.timetable.trees[tree_name]
        rest_tree = create_rest_timetable_tree(self.scheduler.timetable, tree_obj)

        for process_name in tree_obj.process_hierarchy:
            thread_handler = self.scheduler.managed_handlers[process_name]
            rest_process = create_rest_managed_scheduler_entry(thread_handler,
                                                               self.scheduler.timetable,
                                                               self.scheduler.gc)
            rest_tree.processes[process_name] = rest_process.document
        return rest_tree

    @cached_property
    def tree_details(self):
        tree_name = self.request.args.get('tree_name')
        if tree_name:
            return self._get_tree_details(tree_name).document
        else:
            return dict()

    @cached_property
    def mx_page_entries(self):
        resp = dict()
        for tree_name, tree in self.scheduler.timetable.trees.items():
            if tree.mx_page in self.request.path:
                rest_tree = self._get_tree_details(tree_name)
                resp[tree.tree_name] = rest_tree.document
        return resp



__author__ = 'Bohdan Mushkevych'

import pymongo
from synergy.db.manager import ds_manager
from synergy.db.model.freerun_process_entry import ENTRY_NAME
from synergy.db.model.managed_process_entry import PROCESS_NAME, ManagedProcessEntry
from synergy.db.model.unit_of_work import TIMEPERIOD, START_OBJ_ID, END_OBJ_ID

from synergy.db.dao.managed_process_dao import ManagedProcessDao

from synergy.conf import context, settings
from synergy.scheduler.scheduler_constants import *
from synergy.system.data_logging import get_logger


def synch_db():
    """ function reads managed_process and updates context entries appropriately """
    logger = get_logger(PROCESS_SCHEDULER)
    managed_process_dao = ManagedProcessDao(logger)

    try:
        process_entries = managed_process_dao.get_all()
    except LookupError:
        logger.error('Synergy DB is not initialized. Aborting.')
        exit(1)

    for process_entry in process_entries:
        process_name = process_entry.process_name
        if process_name not in context.process_context:
            logger.error('Process {0} has no reflection in the context. Skipping it.'.format(process_name))
            continue

        if not isinstance(context.process_context[process_name], ManagedProcessEntry):
            logger.error('Process entry {0} of non-managed type {1} found in managed_process table. Skipping it.'
                         .format(process_name, context.process_context[process_name].__class__.__name__))
            continue

        context.process_context[process_name] = process_entry
        logger.info('Context updated with process entry {0}.'.format(process_entry.key))


def update_db():
    """ writes to managed_process table records from the context.process_context """
    logger = get_logger(PROCESS_SCHEDULER)
    managed_process_dao = ManagedProcessDao(logger)
    managed_process_dao.clear()

    for process_name, process_entry in context.process_context.items():
        if not isinstance(process_entry, ManagedProcessEntry):
            continue

        managed_process_dao.update(process_entry)
        logger.info('Updated DB with process entry {0} from the context.'.format(process_entry.key))


def reset_db():
    """ drops the *scheduler* database, resets schema """
    logger = get_logger(PROCESS_SCHEDULER)
    logger.info('Starting *scheduler* DB reset')

    ds = ds_manager.ds_factory(logger)
    ds._db_client.drop_database(settings.settings['mongo_db_name'])
    logger.info('*scheduler* db has been dropped')

    connection = ds.connection(COLLECTION_MANAGED_PROCESS)
    connection.create_index([(PROCESS_NAME, pymongo.ASCENDING)], unique=True)

    connection = ds.connection(COLLECTION_FREERUN_PROCESS)
    connection.create_index([(PROCESS_NAME, pymongo.ASCENDING), (ENTRY_NAME, pymongo.ASCENDING)], unique=True)

    connection = ds.connection(COLLECTION_UNIT_OF_WORK)
    connection.create_index([(PROCESS_NAME, pymongo.ASCENDING),
                             (TIMEPERIOD, pymongo.ASCENDING),
                             (START_OBJ_ID, pymongo.ASCENDING),
                             (END_OBJ_ID, pymongo.ASCENDING)], unique=True)

    for collection_name in [COLLECTION_JOB_HOURLY, COLLECTION_JOB_DAILY,
                            COLLECTION_JOB_MONTHLY, COLLECTION_JOB_YEARLY]:
        connection = ds.connection(collection_name)
        connection.create_index([(PROCESS_NAME, pymongo.ASCENDING), (TIMEPERIOD, pymongo.ASCENDING)], unique=True)
    logger.info('*scheduler* db has been recreated')


if __name__ == '__main__':
    pass



__author__ = 'Bohdan Mushkevych'

import heapq
from datetime import datetime, timedelta

from synergy.conf import settings
from synergy.system import time_helper
from synergy.system.time_qualifier import QUALIFIER_REAL_TIME


def compute_release_time(lag_in_minutes):
    future_dt = datetime.utcnow() + timedelta(minutes=lag_in_minutes)
    release_time_str = time_helper.datetime_to_synergy(QUALIFIER_REAL_TIME, future_dt)
    return long(release_time_str)


class PriorityEntry(object):
    """ an entry for Priority Queue, where priority is *entry creation time* + *waiting time* """

    # Creation counter keeps track of PriorityEntry declaration order
    # Each time an instance is created the counter should be increased
    creation_counter = 0

    def __init__(self, entry, lag_in_minutes=settings.settings['gc_release_lag_minutes']):
        """ :param entry: the unit_of_work to reprocess """
        self.entry = entry
        self.release_time = compute_release_time(lag_in_minutes)  # SYNERGY_SESSION_PATTERN: time in the future
        self.creation_counter = PriorityEntry.creation_counter + 1
        PriorityEntry.creation_counter += 1

    def __eq__(self, other):
        return self.entry == other.entry

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        """Defines behavior for the less-than operator, <."""
        if self.release_time == other.release_time:
            return self.creation_counter < other.creation_counter
        else:
            return self.release_time < other.release_time

    def __gt__(self, other):
        """Defines behavior for the greater-than operator, >."""
        if self.release_time == other.release_time:
            return self.creation_counter > other.creation_counter
        else:
            return self.release_time > other.release_time

    def __le__(self, other):
        """Defines behavior for the less-than-or-equal-to operator, <=."""
        if self.release_time == other.release_time:
            return self.creation_counter <= other.creation_counter
        else:
            return self.release_time <= other.release_time

    def __ge__(self, other):
        """Defines behavior for the greater-than-or-equal-to operator, >=."""
        if self.release_time == other.release_time:
            return self.creation_counter >= other.creation_counter
        else:
            return self.release_time >= other.release_time

    def __hash__(self):
        return hash((self.release_time, self.creation_counter))


class PriorityQueue(object):
    """ Priority Queue that retrieves entries in the priority order (lowest first) """

    def __init__(self):
        self.queue = list()

    def __len__(self):
        return len(self.queue)

    def __contains__(self, item):
        return item in self.queue

    def put(self, item):
        heapq.heappush(self.queue, item)

    def pop(self):
        """ :return: minimal element is removed from the queue and returned to the caller """
        return heapq.heappop(self.queue)

    def peek(self):
        """ :return: minimal element without being removed from the queue """
        return min(self.queue)



__author__ = 'Bohdan Mushkevych'

from collections import OrderedDict

from synergy.conf import context
from synergy.db.model.managed_process_entry import ManagedProcessEntry
from synergy.system.time_qualifier import *
from synergy.system.time_helper import cast_to_time_qualifier
from synergy.system.timeperiod_dict import TimeperiodDict


class HierarchyEntry(object):
    def __init__(self, hierarchy, parent, process_entry):
        assert isinstance(process_entry, ManagedProcessEntry)
        self.hierarchy = hierarchy
        self.parent = parent
        self.process_entry = process_entry
        self.timeperiod_dict = TimeperiodDict(process_entry.time_qualifier, process_entry.time_grouping)

    def cast_timeperiod(self, timeperiod):
        return cast_to_time_qualifier(self.process_entry.time_qualifier, timeperiod)


class ProcessHierarchy(object):
    def __init__(self, *process_names):
        self.entries = OrderedDict()
        self.qualifiers = OrderedDict()

        top_node = None
        process_entries = [context.process_context[process_name] for process_name in process_names]

        sorted_process_entries = sorted(process_entries, key=lambda x: QUALIFIER_DICT[x.time_qualifier], reverse=True)
        for process_entry in sorted_process_entries:
            assert isinstance(process_entry, ManagedProcessEntry)
            entry = HierarchyEntry(self, top_node, process_entry)
            top_node = entry
            self.entries[process_entry.process_name] = entry
            self.qualifiers[process_entry.time_qualifier] = entry

    def __contains__(self, value):
        """
        :param value: process name
        :return: True if a hierarchy entry for the process_entry with the given name is registered in this hierarchy;
                 False otherwise
        """
        return value in self.entries

    def __getitem__(self, key):
        """
        :param key: process name
        :return: associated hierarchy entry of HierarchyEntry type
        """
        return self.entries[key]

    def __iter__(self):
        """ for x in self
        :return process_names in descending order of their time_qualifier: yearly->monthly->daily->hourly """
        return iter(self.entries)

    def __str__(self):
        msg = 'Process Hierarchy: '
        for process_name, hierarchy_entry in self.entries:
            msg += '{1}->{0} '.format(process_name, hierarchy_entry.process_entry.time_qualifier)
        return msg

    def has_qualifier(self, qualifier):
        """
        :param qualifier: time_qualifier
        :return: True if a HierarchyEntry with given time_qualifier is registered in this hierarchy; False otherwise
        """
        return qualifier in self.qualifiers

    def get_by_qualifier(self, qualifier):
        """
        :param qualifier: time_qualifier of the searched process
        :return: associated entry of HierarchyEntry type
                or None if no process with given time_qualifier is registered in this hierarchy
        """
        return self.qualifiers.get(qualifier, None)

    def get_child_by_qualifier(self, parent_qualifier):
        """
        :param parent_qualifier: time_qualifier of the parent process
        :return: <HierarchyEntry> child entry to the HierarchyEntry associated with the parent_qualifier
                or None if the given parent_qualifier is not registered in this hierarchy
                or None if the given parent_qualifier is the bottom process
        """
        if parent_qualifier not in self.qualifiers:
            return None

        process_qualifiers = self.qualifiers.keys()
        if parent_qualifier == process_qualifiers[-1]:
            return None

        parent_index = process_qualifiers.index(parent_qualifier)
        return self.qualifiers[process_qualifiers[parent_index + 1]]

    @property
    def top_process(self):
        """ :return: <ManagedProcessEntry> of the hierarchy's top entry """
        key = next(iter(self.entries))
        return self.entries[key].process_entry

    @property
    def bottom_process(self):
        """ :return: <ManagedProcessEntry> of the hierarchy's bottom entry """
        key = next(reversed(self.entries))
        return self.entries[key].process_entry



__author__ = 'Bohdan Mushkevych'

from pymongo import MongoClient, ASCENDING, DESCENDING
from bson.objectid import ObjectId

from synergy.conf import settings
from synergy.db.model.unit_of_work import TIMEPERIOD

QUERY_GET_ALL = {}

if 'ds_factory' not in globals():
    # this block defines global variable ds_factory

    def factory():
        # the only way to implement nonlocal closure variables in Python 2.X
        instances = {}

        def get_instance(logger):
            ds_type = settings.settings['ds_type']

            if ds_type not in instances:
                if ds_type == "mongo_db":
                    instances[ds_type] = MongoDbManager(logger)
                elif ds_type == "hbase":
                    instances[ds_type] = HBaseManager(logger)
                else:
                    raise ValueError('Unsupported Data Source type: {0}'.format(ds_type))
            return instances[ds_type]

        return get_instance

    global ds_factory
    ds_factory = factory()


class BaseManager(object):
    """
    BaseManager holds definition of the Data Source and an interface to read, write, delete and update (CRUD)
    models withing the DataSource
    """

    def __init__(self, logger):
        super(BaseManager, self).__init__()
        self.logger = logger

    def __str__(self):
        raise NotImplementedError('method __str__ must be implemented by {0}'.format(self.__class__.__name__))

    def is_alive(self):
        """ :return: True if the database server is available. False otherwise """
        raise NotImplementedError('method is_alive must be implemented by {0}'.format(self.__class__.__name__))

    def get(self, table_name, primary_key):
        raise NotImplementedError('method get must be implemented by {0}'.format(self.__class__.__name__))

    def filter(self, table_name, query):
        raise NotImplementedError('method filter must be implemented by {0}'.format(self.__class__.__name__))

    def update(self, table_name, primary_key, instance):
        raise NotImplementedError('method update must be implemented by {0}'.format(self.__class__.__name__))

    def delete(self, table_name, primary_key):
        raise NotImplementedError('method delete must be implemented by {0}'.format(self.__class__.__name__))

    def highest_primary_key(self, table_name, timeperiod_low, timeperiod_high):
        raise NotImplementedError(
            'method highest_primary_key must be implemented by {0}'.format(self.__class__.__name__))

    def lowest_primary_key(self, table_name, timeperiod_low, timeperiod_high):
        raise NotImplementedError(
            'method lowest_primary_key must be implemented by {0}'.format(self.__class__.__name__))

    def cursor_fine(self,
                    table_name,
                    start_id_obj,
                    end_id_obj,
                    iteration,
                    start_timeperiod,
                    end_timeperiod):
        """ method returns DB cursor based on precise boundaries """
        raise NotImplementedError('method cursor_fine must be implemented by {0}'.format(self.__class__.__name__))

    def cursor_batch(self,
                     table_name,
                     start_timeperiod,
                     end_timeperiod):
        """ method returns batched DB cursor """
        raise NotImplementedError('method cursor_batch must be implemented by {0}'.format(self.__class__.__name__))


class MongoDbManager(BaseManager):
    def __init__(self, logger):
        super(MongoDbManager, self).__init__(logger)
        self._db_client = MongoClient(settings.settings['mongodb_host_list'])
        self._db = self._db_client[settings.settings['mongo_db_name']]

    def __del__(self):
        try:
            self._db_client.close()
        except AttributeError:
            pass

    def __str__(self):
        return 'MongoDbManager: {0}@{1}' % (settings.settings['mongodb_host_list'], settings.settings['mongo_db_name'])

    def is_alive(self):
        return self._db_client.alive()

    def connection(self, table_name):
        return self._db[table_name]

    def filter(self, table_name, query):
        conn = self._db[table_name]
        return conn.find(query)

    def delete(self, table_name, primary_key):
        conn = self._db[table_name]
        return conn.remove(primary_key, safe=True)

    def get(self, table_name, primary_key):
        query = {'_id': primary_key}
        conn = self._db[table_name]
        db_entry = conn.find_one(query)
        if db_entry is None:
            msg = 'Instance with ID={0} was not found'.format(primary_key)
            self.logger.warn(msg)
            raise LookupError(msg)
        return db_entry

    def insert(self, table_name, instance):
        conn = self._db[table_name]
        return conn.insert(instance, safe=True)

    def update(self, table_name, primary_key, instance):
        conn = self._db[table_name]
        conn.update(primary_key, instance, upsert=True, safe=True)

    def highest_primary_key(self, table_name, timeperiod_low, timeperiod_high):
        query = {TIMEPERIOD: {'$gte': timeperiod_low, '$lt': timeperiod_high}}
        conn = self._db[table_name]
        asc_search = conn.find(spec=query, fields='_id').sort('_id', ASCENDING).limit(1)
        if asc_search.count() == 0:
            raise LookupError('No records in timeperiod: [{0} : {1}) in collection {2}'
                              .format(timeperiod_low, timeperiod_high, table_name))
        return asc_search[0]['_id']

    def lowest_primary_key(self, table_name, timeperiod_low, timeperiod_high):
        query = {TIMEPERIOD: {'$gte': timeperiod_low, '$lt': timeperiod_high}}
        conn = self._db[table_name]
        dec_search = conn.find(spec=query, fields='_id').sort('_id', DESCENDING).limit(1)
        if dec_search.count() == 0:
            raise LookupError('No records in timeperiod: [{0} : {1}) in collection {2}'
                              .format(timeperiod_low, timeperiod_high, table_name))
        return dec_search[0]['_id']

    def cursor_fine(self,
                    table_name,
                    start_id_obj,
                    end_id_obj,
                    iteration,
                    start_timeperiod,
                    end_timeperiod):
        if not isinstance(start_id_obj, ObjectId):
            start_id_obj = ObjectId(start_id_obj)
        if not isinstance(end_id_obj, ObjectId):
            end_id_obj = ObjectId(end_id_obj)

        if iteration == 0:
            queue = {'_id': {'$gte': start_id_obj, '$lte': end_id_obj}}
        else:
            queue = {'_id': {'$gt': start_id_obj, '$lte': end_id_obj}}

        if start_timeperiod is not None and end_timeperiod is not None:
            # remove all accident objects that may be in [start_id_obj : end_id_obj] range
            queue[TIMEPERIOD] = {'$gte': start_timeperiod, '$lt': end_timeperiod}

        conn = self._db[table_name]
        batch_size = settings.settings['batch_size']
        return conn.find(queue).sort('_id', ASCENDING).limit(batch_size)

    def cursor_batch(self, table_name, start_timeperiod, end_timeperiod):
        assert start_timeperiod is not None and end_timeperiod is not None

        conn = self._db[table_name]
        batch_size = settings.settings['batch_size']

        queue = {TIMEPERIOD: {'$gte': start_timeperiod, '$lt': end_timeperiod}}
        return conn.find(queue).batch_size(batch_size)


class HBaseManager(BaseManager):
    pass



__author__ = 'Bohdan Mushkevych'

from synergy.db.dao.base_dao import BaseDao
from synergy.db.model.box_configuration import BoxConfiguration, BOX_ID, PROCESS_NAME
from synergy.supervisor.supervisor_constants import COLLECTION_BOX_CONFIGURATION

QUERY_PROCESSES_FOR_BOX_ID = lambda box_id: {BOX_ID: box_id}


class BoxConfigurationDao(BaseDao):
    """ Thread-safe Data Access Object for box_configuration table/collection """
    def __init__(self, logger):
        super(BoxConfigurationDao, self).__init__(logger=logger,
                                                  model_class=BoxConfiguration,
                                                  primary_key=[BOX_ID, PROCESS_NAME],
                                                  collection_name=COLLECTION_BOX_CONFIGURATION)



__author__ = 'Bohdan Mushkevych'

from logging import ERROR

from synergy.db.model import job
from synergy.scheduler.scheduler_constants import STATE_MACHINE_CONTINUOUS
from synergy.scheduler.abstract_state_machine import AbstractStateMachine
from synergy.system import time_helper
from synergy.conf import context


class StateMachineContinuous(AbstractStateMachine):
    """ State Machine of 6 states is expected to spawn multiple UOWs per job/timeperiod.
        Job timeperiods boundaries are meant to be discrete/fixed
        in comparison to StateMachineRecomputing this one does not re-compute processing boundaries """

    def __init__(self, logger, timetable, name=STATE_MACHINE_CONTINUOUS):
        super(StateMachineContinuous, self).__init__(logger, timetable, name)

    def __del__(self):
        super(StateMachineContinuous, self).__del__()

    @property
    def run_on_active_timeperiod(self):
        """ :return: True, as we allow multiple runs on a given timeperiod """
        return True

    def notify(self, uow):
        tree = self.timetable.get_tree(uow.process_name)
        node = tree.get_node(uow.process_name, uow.timeperiod)
        job_record = node.job_record

        if not job_record.is_final_run:
            self.logger.info('Suppressing state change for Job {0}@{1}, since it is not in STATE_FINAL_RUN'
                             .format(uow.process_name, uow.timeperiod))
            return
        self._process_state_final_run(job_record)

    def _compute_next_job_state(self, job_record):
        time_qualifier = context.process_context[job_record.process_name].time_qualifier
        actual_timeperiod = time_helper.actual_timeperiod(time_qualifier)
        is_job_finalizable = self.timetable.is_job_record_finalizable(job_record)

        if job_record.timeperiod == actual_timeperiod or is_job_finalizable is False:
            return job.STATE_IN_PROGRESS

        elif job_record.timeperiod < actual_timeperiod and is_job_finalizable is True:
            return job.STATE_FINAL_RUN

        else:
            msg = 'Job record {0} has timeperiod from future {1} vs current time {2}' \
                .format(job_record.db_id, job_record.timeperiod, actual_timeperiod)
            self._log_message(ERROR, job_record.process_name, job_record.timeperiod, msg)
            raise ValueError(msg)

    def _process_state_embryo(self, job_record):
        """ method that takes care of processing job records in STATE_EMBRYO state"""
        time_qualifier = context.process_context[job_record.process_name].time_qualifier
        end_timeperiod = time_helper.increment_timeperiod(time_qualifier, job_record.timeperiod)
        uow, is_duplicate = self.insert_and_publish_uow(job_record.process_name,
                                                        job_record.timeperiod,
                                                        end_timeperiod,
                                                        0,
                                                        0)

        try:
            target_state = self._compute_next_job_state(job_record)
            self.update_job(job_record, uow, target_state)
        except ValueError:
            # do no processing for the future timeperiods
            pass

    def _process_state_in_progress(self, job_record):
        """ method that takes care of processing job records in STATE_IN_PROGRESS state"""
        def _process_state(target_state, uow):
            if uow.is_active:
                # Large Job processing takes more than 1 tick of the Scheduler
                # Let the Job processing complete - do no updates to Scheduler records
                pass
            elif uow.is_finished:
                # create new UOW to cover new inserts
                new_uow, is_duplicate = self.insert_and_publish_uow(job_record.process_name,
                                                                    job_record.timeperiod,
                                                                    end_timeperiod,
                                                                    0,
                                                                    int(uow.end_id) + 1)
                self.update_job(job_record, new_uow, target_state)

        time_qualifier = context.process_context[job_record.process_name].time_qualifier
        end_timeperiod = time_helper.increment_timeperiod(time_qualifier, job_record.timeperiod)
        uow = self.uow_dao.get_one(job_record.related_unit_of_work)

        try:
            target_state = self._compute_next_job_state(job_record)
            _process_state(target_state, uow)
        except ValueError:
            # do no processing for the future timeperiods
            pass



__author__ = 'Bohdan Mushkevych'

import time
import os

import psutil

from synergy.system.repeat_timer import RepeatTimer
from synergy.conf import settings


class FootprintCalculator(object):
    def __init__(self):
        self.pid = os.getpid()

    def group(self, number):
        """ method formats number and inserts thousands separators """
        s = str(number)
        groups = []
        while s and s[-1].isdigit():
            groups.append(s[-3:])
            s = s[:-3]
        return s + '\''.join(reversed(groups))

    @property
    def document(self):
        ps = psutil.Process(self.pid)
        return {'memory_rss': self.group(ps.get_memory_info()[0]),
                'memory_vms': self.group(ps.get_memory_info()[1]),
                'cpu_utilization': '%02d' % ps.get_cpu_percent(),
                'mem_virtual_free': self.group(psutil.virtual_memory().free),
                'mem_swap_free': self.group(psutil.swap_memory().free)}

    def get_snapshot(self):
        resp = 'Footprint: RSS={memory_rss} VMS={memory_vms} CPU={cpu_utilization}; ' \
               'Available: PHYS={mem_virtual_free} VIRT={mem_swap_free}'.format(**self.document)
        return resp


class Tracker(object):
    def __init__(self, name):
        self.name = name
        self.per_24h = 0
        self.per_tick = 0

    def increment(self, delta=1):
        self.per_24h += delta
        self.per_tick += delta

    def reset_tick(self):
        self.per_tick = 0

    def reset_24h(self):
        self.per_24h = 0


class TrackerPair(object):
    def __init__(self, name, success='Success', failure='Failure'):
        self.name = name
        self.success = Tracker(success)
        self.failure = Tracker(failure)

    def increment_success(self, delta=1):
        self.success.increment(delta)

    def increment_failure(self, delta=1):
        self.failure.increment(delta)

    def reset_tick(self):
        self.success.reset_tick()
        self.failure.reset_tick()

    def reset_24h(self):
        self.success.reset_24h()
        self.failure.reset_24h()

    def to_string(self, tick_interval_seconds, show_header=True):
        header = self.name + ' : ' + self.success.name + '/' + self.failure.name + '.' if show_header else ''
        return header + 'In last {0:d} seconds: {1:d}/{2:d}. In last 24 hours: {3:d}/{4:d}'.format(
            tick_interval_seconds,
            self.success.per_tick,
            self.failure.per_tick,
            self.success.per_24h,
            self.failure.per_24h)


class TickerThread(object):
    SECONDS_IN_24_HOURS = 86400
    TICKS_BETWEEN_FOOTPRINTS = 10

    def __init__(self, logger):
        self.logger = logger
        self.trackers = dict()
        self.interval = settings.settings['perf_ticker_interval']
        self.mark_24_hours = time.time()
        self.mark_footprint = time.time()
        self.footprint = FootprintCalculator()
        self.timer = RepeatTimer(self.interval, self._run_tick_thread)

    def add_tracker(self, tracker):
        self.trackers[tracker.name] = tracker

    def get_tracker(self, name):
        return self.trackers[name]

    def start(self):
        self.timer.start()

    def cancel(self):
        self.timer.cancel()

    def is_alive(self):
        return self.timer.is_alive()

    def _print_footprint(self):
        if time.time() - self.mark_footprint > self.TICKS_BETWEEN_FOOTPRINTS * self.interval:
            self.logger.info(self.footprint.get_snapshot())
            self.mark_footprint = time.time()

    def _run_tick_thread(self):
        self._print_footprint()

        current_time = time.time()
        do_24h_reset = current_time - self.mark_24_hours > self.SECONDS_IN_24_HOURS
        if do_24h_reset:
            self.mark_24_hours = current_time

        tracker_outputs = []
        for tracker_name, tracker in self.trackers.items():
            tracker_outputs.append(tracker.to_string(self.interval))
            tracker.reset_tick()
            if do_24h_reset:
                tracker.reset_24h()

        self.logger.info('\n'.join(tracker_outputs))


class SimpleTracker(TickerThread):
    TRACKER_PERFORMANCE = 'Performance'

    def __init__(self, logger):
        super(SimpleTracker, self).__init__(logger)
        self.add_tracker(TrackerPair(self.TRACKER_PERFORMANCE))

    @property
    def tracker(self):
        return self.get_tracker(self.TRACKER_PERFORMANCE)


class SessionPerformanceTracker(TickerThread):
    TRACKER_INSERT = 'Insert'
    TRACKER_UPDATE = 'Update'

    def __init__(self, logger):
        super(SessionPerformanceTracker, self).__init__(logger)
        self.add_tracker(TrackerPair(self.TRACKER_INSERT))
        self.add_tracker(TrackerPair(self.TRACKER_UPDATE))

    @property
    def insert(self):
        return self.get_tracker(self.TRACKER_INSERT)

    @property
    def update(self):
        return self.get_tracker(self.TRACKER_UPDATE)


class UowAwareTracker(SimpleTracker):
    STATE_IDLE = 'state_idle'
    STATE_PROCESSING = 'state_processing'

    def __init__(self, logger):
        super(UowAwareTracker, self).__init__(logger)
        self.state = self.STATE_IDLE
        self.success_per_job = 0
        self.failure_per_job = 0
        self.uow = None
        self.state_triggered_at = time.time()

    def _run_tick_thread(self):
        super(UowAwareTracker, self)._run_tick_thread()

        if self.state == self.STATE_PROCESSING:
            msg = 'State: {0} for {1:.2f} sec; Success/Failure {2}/{3} in this uow;'\
                  .format(self.state, time.time() - self.state_triggered_at, self.success_per_job, self.failure_per_job)
        else:
            msg = 'State: {0} for {1:.2f} sec;'.format(self.state, time.time() - self.state_triggered_at)
        self.logger.info(msg)

    def increment_success(self):
        self.tracker.increment_success()
        self.success_per_job += 1

    def increment_failure(self):
        self.tracker.increment_failure()
        self.failure_per_job += 1

    def start_uow(self, uow):
        self.state = self.STATE_PROCESSING
        self.uow = uow
        self.state_triggered_at = time.time()

    def finish_uow(self):
        self.logger.info('uow {0} at {1}: Success/Failure {2}/{3} entries in {4:.2f} seconds'.
                         format(self.uow.db_id, self.uow.timeperiod,
                                self.success_per_job, self.failure_per_job, time.time() - self.state_triggered_at))
        self.cancel_uow()

    def cancel_uow(self):
        self.state = self.STATE_IDLE
        self.uow = None
        self.state_triggered_at = time.time()
        self.success_per_job = 0



__author__ = 'Bohdan Mushkevych'

from logging import ERROR, INFO

from synergy.db.model import job
from synergy.db.manager import ds_manager
from synergy.conf import context
from synergy.system.decorator import with_reconnect
from synergy.system import time_helper
from synergy.scheduler.scheduler_constants import STATE_MACHINE_RECOMPUTING
from synergy.scheduler.abstract_state_machine import AbstractStateMachine


class StateMachineRecomputing(AbstractStateMachine):
    """ State Machine of 6 states is expected to spawn multiple UOWs per job/timeperiod.
        Job timeperiods boundaries are dynamic - i.e. underlying data volume grows with time """

    def __init__(self, logger, timetable):
        super(StateMachineRecomputing, self).__init__(logger, timetable, name=STATE_MACHINE_RECOMPUTING)
        self.ds = ds_manager.ds_factory(self.logger)

    def __del__(self):
        super(StateMachineRecomputing, self).__del__()

    @property
    def run_on_active_timeperiod(self):
        """ :return: True, as this StateMachine allows multiple runs for some given timeperiod """
        return True

    def notify(self, uow):
        tree = self.timetable.get_tree(uow.process_name)
        node = tree.get_node(uow.process_name, uow.timeperiod)
        job_record = node.job_record

        if not job_record.is_final_run:
            self.logger.info('Suppressing state change for Job {0}@{1}, since it is not in STATE_FINAL_RUN'
                             .format(uow.process_name, uow.timeperiod))
            return
        self._process_state_final_run(job_record)

    @with_reconnect
    def update_scope_of_processing(self, process_name, uow, start_timeperiod, end_timeperiod):
        """method reads collection and refine slice upper bound for processing"""
        source_collection_name = uow.source
        last_object_id = self.ds.lowest_primary_key(source_collection_name, start_timeperiod, end_timeperiod)
        uow.end_id = str(last_object_id)
        self.uow_dao.update(uow)

        msg = 'Updated processing range for {0}@{1} for collection {2}: [{3} : {4}]' \
              .format(process_name, start_timeperiod, source_collection_name, uow.start_id, uow.end_id)
        self._log_message(INFO, process_name, start_timeperiod, msg)

    def _compute_and_transfer_to_progress(self, process_name, start_timeperiod, end_timeperiod, job_record):
        """ method computes new unit_of_work for job record in STATE_IN_PROGRESS
        it also contains _fuzzy_ logic regard the DuplicateKeyError:
        - we try to compute new scope of processing
        - in case we face DuplicateKeyError, we try to recover from it by reading existing unit_of_work from DB:
        -- in case unit_of_work can be located - we update job record and proceed normally
        -- in case unit_of_work can not be located (what is equal to fatal data corruption) - we log exception and
        ask/expect manual intervention to resolve the corruption"""
        source_collection_name = context.process_context[process_name].source
        start_id = self.ds.highest_primary_key(source_collection_name, start_timeperiod, end_timeperiod)
        end_id = self.ds.lowest_primary_key(source_collection_name, start_timeperiod, end_timeperiod)
        uow, is_duplicate = self.insert_and_publish_uow(process_name,
                                                        start_timeperiod,
                                                        end_timeperiod,
                                                        start_id,
                                                        end_id)
        self.update_job(job_record, uow, job.STATE_IN_PROGRESS)

    def _compute_and_transfer_to_final_run(self, process_name, start_timeperiod, end_timeperiod, job_record):
        """ method computes new unit_of_work and transfers the job to STATE_FINAL_RUN
        it also shares _fuzzy_ DuplicateKeyError logic from _compute_and_transfer_to_progress method"""
        source_collection_name = context.process_context[process_name].source
        start_id = self.ds.highest_primary_key(source_collection_name, start_timeperiod, end_timeperiod)
        end_id = self.ds.lowest_primary_key(source_collection_name, start_timeperiod, end_timeperiod)
        uow, transfer_to_final = self.insert_and_publish_uow(process_name,
                                                             start_timeperiod,
                                                             end_timeperiod,
                                                             start_id,
                                                             end_id)
        self.update_job(job_record, uow, job.STATE_FINAL_RUN)

        if transfer_to_final:
            self._process_state_final_run(job_record)

    def _process_state_embryo(self, job_record):
        """ method that takes care of processing job records in STATE_EMBRYO state"""
        time_qualifier = context.process_context[job_record.process_name].time_qualifier
        end_timeperiod = time_helper.increment_timeperiod(time_qualifier, job_record.timeperiod)
        self._compute_and_transfer_to_progress(job_record.process_name, job_record.timeperiod,
                                               end_timeperiod, job_record)

    def _process_state_in_progress(self, job_record):
        """ method that takes care of processing job records in STATE_IN_PROGRESS state"""
        time_qualifier = context.process_context[job_record.process_name].time_qualifier
        end_timeperiod = time_helper.increment_timeperiod(time_qualifier, job_record.timeperiod)
        actual_timeperiod = time_helper.actual_timeperiod(time_qualifier)
        is_job_finalizable = self.timetable.is_job_record_finalizable(job_record)
        uow = self.uow_dao.get_one(job_record.related_unit_of_work)

        if job_record.timeperiod == actual_timeperiod or is_job_finalizable is False:
            if uow.is_invalid or uow.is_requested:
                # current uow has not been processed yet. update it
                self.update_scope_of_processing(job_record.process_name, uow, job_record.timeperiod, end_timeperiod)
            else:
                # STATE_IN_PROGRESS, STATE_PROCESSED, STATE_CANCELED, STATE_NOOP
                # create new uow to cover new inserts
                self._compute_and_transfer_to_progress(job_record.process_name, job_record.timeperiod,
                                                       end_timeperiod, job_record)

        elif job_record.timeperiod < actual_timeperiod and is_job_finalizable is True:
            # create new uow for FINAL RUN
            self._compute_and_transfer_to_final_run(job_record.process_name, job_record.timeperiod,
                                                    end_timeperiod, job_record)

        else:
            msg = 'Job {0} has timeperiod from future {1} vs current time {2}' \
                  .format(job_record.db_id, job_record.timeperiod, actual_timeperiod)
            self._log_message(ERROR, job_record.process_name, job_record.timeperiod, msg)



__author__ = 'Bohdan Mushkevych'

import json

from synergy.db.model.freerun_process_entry import FreerunProcessEntry
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao
from synergy.db.dao.freerun_process_dao import FreerunProcessDao
from synergy.mx.base_request_handler import valid_action_request
from synergy.mx.abstract_action_handler import AbstractActionHandler
from synergy.scheduler.scheduler_constants import STATE_MACHINE_FREERUN


class FreerunActionHandler(AbstractActionHandler):
    def __init__(self, request, **values):
        super(FreerunActionHandler, self).__init__(request, **values)
        self.process_name = self.request_arguments.get('process_name')
        self.entry_name = self.request_arguments.get('entry_name')
        self.freerun_process_dao = FreerunProcessDao(self.logger)
        self.uow_dao = UnitOfWorkDao(self.logger)
        self.is_request_valid = True if self.process_name and self.entry_name else False

        if self.is_request_valid:
            self.process_name = self.process_name.strip()
            self.entry_name = self.entry_name.strip()
            self.is_requested_state_on = 'is_on' in self.request_arguments and self.request_arguments['is_on']

    @AbstractActionHandler.thread_handler.getter
    def thread_handler(self):
        handler_key = (self.process_name, self.entry_name)
        return self.scheduler.freerun_handlers[handler_key]

    @AbstractActionHandler.process_entry.getter
    def process_entry(self):
        return self.thread_handler.process_entry

    @valid_action_request
    def action_cancel_uow(self):
        freerun_state_machine = self.scheduler.timetable.state_machines[STATE_MACHINE_FREERUN]
        freerun_state_machine.cancel_uow(self.process_entry)
        return self.reply_ok()

    @valid_action_request
    def action_get_uow(self):
        uow_id = self.process_entry.related_unit_of_work
        if uow_id is None:
            resp = {'response': 'no related unit_of_work'}
        else:
            resp = self.uow_dao.get_one(uow_id).document
            for key in resp:
                resp[key] = str(resp[key])
        return resp

    @valid_action_request
    def action_get_log(self):
        return {'log': self.process_entry.log}

    @valid_action_request
    def action_update_entry(self):
        if 'insert_button' in self.request_arguments:
            process_entry = FreerunProcessEntry()
            process_entry.process_name = self.process_name
            process_entry.entry_name = self.entry_name

            if self.request_arguments['arguments']:
                arguments = self.request_arguments['arguments'].decode('unicode-escape')
                process_entry.arguments = json.loads(arguments)
            else:
                process_entry.arguments = {}

            process_entry.description = self.request_arguments['description']
            process_entry.is_on = self.is_requested_state_on
            process_entry.trigger_frequency = self.request_arguments['trigger_frequency']
            self.freerun_process_dao.update(process_entry)

            self.scheduler._register_process_entry(process_entry, self.scheduler.fire_freerun_worker)

        elif 'update_button' in self.request_arguments:
            is_interval_changed = self.process_entry.trigger_frequency != self.request_arguments['trigger_frequency']

            if self.request_arguments['arguments']:
                arguments = self.request_arguments['arguments'].decode('unicode-escape')
                self.process_entry.arguments = json.loads(arguments)
            else:
                self.process_entry.arguments = {}

            self.process_entry.description = self.request_arguments['description']
            self.process_entry.is_on = self.is_requested_state_on
            self.process_entry.trigger_frequency = self.request_arguments['trigger_frequency']
            self.freerun_process_dao.update(self.process_entry)

            if is_interval_changed:
                self.action_change_interval()

            if self.process_entry.is_on != self.is_requested_state_on:
                if self.is_requested_state_on:
                    self.action_activate_trigger()
                else:
                    self.action_deactivate_trigger()

        elif 'delete_button' in self.request_arguments:
            handler_key = (self.process_name, self.entry_name)
            self.thread_handler.deactivate()
            self.freerun_process_dao.remove(handler_key)
            del self.scheduler.freerun_handlers[handler_key]

        elif 'cancel_button' in self.request_arguments:
            pass

        else:
            self.logger.error('MX Error: unknown action requested by schedulable_form.html')

        return self.reply_ok()



__author__ = 'Bohdan Mushkevych'

from synergy.db.dao.freerun_process_dao import FreerunProcessDao
from synergy.db.dao.managed_process_dao import ManagedProcessDao
from synergy.system.event_clock import parse_time_trigger_string


class ThreadHandlerHeader(object):
    """ ThreadHandlerHeader is a data structure representing key Thread Handler features.
        It is passed to the Timer instance and later on - to the Scheduler's running function as an argument """

    def __init__(self, key, trigger_frequency, process_entry):
        self.key = key
        self.trigger_frequency = trigger_frequency
        self.process_entry = process_entry


class AbstractThreadHandler(object):
    """ ThreadHandler is a thread running within the Synergy Scheduler and triggering Scheduler's fire_XXX logic"""

    def __init__(self, logger, key, trigger_frequency, call_back, process_entry):
        self.logger = logger
        self.key = key
        self.trigger_frequency = trigger_frequency
        self.call_back = call_back
        self.process_entry = process_entry

        parsed_trigger_frequency, timer_klass = parse_time_trigger_string(trigger_frequency)
        self.timer_instance = timer_klass(parsed_trigger_frequency, call_back, args=[self.header])
        self.is_started = False
        self.is_terminated = False

        self.logger.info('Created {0} for {1}~{2}'.format(self.__class__.__name__, key, trigger_frequency))

    def __del__(self):
        self.timer_instance.cancel()

    @property
    def header(self):
        return ThreadHandlerHeader(self.key, self.trigger_frequency, self.process_entry)

    @property
    def dao(self):
        raise NotImplementedError('property dao must be implemented by {0}'.format(self.__class__.__name__))

    def activate(self, update_persistent=True):
        if self.timer_instance.is_alive():
            return

        if self.is_terminated:
            parsed_trigger_frequency, timer_klass = parse_time_trigger_string(self.trigger_frequency)
            self.timer_instance = timer_klass(parsed_trigger_frequency, self.call_back, args=[self.header])

        self.process_entry.is_on = True
        if update_persistent:
            self.dao.update(self.process_entry)

        self.timer_instance.start()
        self.is_terminated = False
        self.is_started = True

    def deactivate(self, update_persistent=True):
        self.timer_instance.cancel()
        self.is_terminated = True

        self.process_entry.is_on = False
        if update_persistent:
            self.dao.update(self.process_entry)

    def trigger(self):
        self.timer_instance.trigger()

    def change_interval(self, value, update_persistent=True):
        parsed_trigger_frequency, timer_klass = parse_time_trigger_string(value)

        if isinstance(self.timer_instance, timer_klass):
            # trigger time has changed only frequency of run
            self.timer_instance.change_interval(parsed_trigger_frequency)
        else:
            # trigger time requires different type of timer - RepeatTimer instead of EventClock or vice versa
            # 1. deactivate current timer
            self.deactivate()

            # 2. create a new timer instance
            parsed_trigger_frequency, timer_klass = parse_time_trigger_string(self.trigger_frequency)
            self.timer_instance = timer_klass(parsed_trigger_frequency, self.call_back, args=[self.header])
            self.trigger_frequency = value

            # 3. start if necessary
            if self.process_entry.is_on:
                self.timer_instance.start()
                self.is_terminated = False
                self.is_started = True

        self.process_entry.trigger_frequency = value
        if update_persistent:
            self.dao.update(self.process_entry)

    def next_run_in(self, utc_now=None):
        return self.timer_instance.next_run_in(utc_now)

    @property
    def is_alive(self):
        return self.timer_instance.is_alive()


class FreerunThreadHandler(AbstractThreadHandler):
    def __init__(self, logger, key, trigger_frequency, call_back, process_entry):
        super(FreerunThreadHandler, self).__init__(logger, key, trigger_frequency, call_back, process_entry)
        self.freerun_process_dao = FreerunProcessDao(self.logger)

    @property
    def dao(self):
        return self.freerun_process_dao


class ManagedThreadHandler(AbstractThreadHandler):
    def __init__(self, logger, key, trigger_frequency, call_back, process_entry):
        super(ManagedThreadHandler, self).__init__(logger, key, trigger_frequency, call_back, process_entry)
        self.managed_process_dao = ManagedProcessDao(self.logger)

    @property
    def dao(self):
        return self.managed_process_dao



__author__ = 'Bohdan Mushkevych'

from synergy.db.model import unit_of_work
from synergy.workers.abstract_uow_aware_worker import AbstractUowAwareWorker


class IdentityWorker(AbstractUowAwareWorker):
    """ Marks any incoming unit_of_work as <complete> """

    def __init__(self, process_name):
        super(IdentityWorker, self).__init__(process_name)

    def __del__(self):
        super(IdentityWorker, self).__del__()

    def _process_uow(self, uow):
        return 0, unit_of_work.STATE_PROCESSED



__author__ = 'Bohdan Mushkevych'

from werkzeug.utils import cached_property

from synergy.mx.base_request_handler import BaseRequestHandler
from synergy.mx.rest_model_factory import create_rest_managed_scheduler_entry, create_rest_freerun_scheduler_entry
from synergy.system.performance_tracker import FootprintCalculator


# Scheduler Entries Details tab
class SchedulerEntries(BaseRequestHandler):
    def __init__(self, request, **values):
        super(SchedulerEntries, self).__init__(request, **values)

    @cached_property
    def managed_entries(self):
        list_of_rows = []
        try:
            sorter_keys = sorted(self.scheduler.managed_handlers.keys())
            for key in sorter_keys:
                thread_handler = self.scheduler.managed_handlers[key]
                rest_model = create_rest_managed_scheduler_entry(thread_handler,
                                                                 self.scheduler.timetable,
                                                                 self.scheduler.gc)
                list_of_rows.append(rest_model.document)
        except Exception as e:
            self.logger.error('MX Exception {0}'.format(e), exc_info=True)

        return list_of_rows

    @cached_property
    def freerun_entries(self):
        list_of_rows = []
        try:
            sorter_keys = sorted(self.scheduler.freerun_handlers.keys())
            for key in sorter_keys:
                thread_handler = self.scheduler.freerun_handlers[key]
                rest_model = create_rest_freerun_scheduler_entry(thread_handler)
                list_of_rows.append(rest_model.document)
        except Exception as e:
            self.logger.error('MX Exception {0}'.format(e), exc_info=True)

        return list_of_rows

    @cached_property
    def footprint(self):
        try:
            calculator = FootprintCalculator()
            return calculator.document
        except Exception as e:
            self.logger.error('MX Exception {0}'.format(e), exc_info=True)
            return []

    @cached_property
    def reprocess_uows(self):
        return self.scheduler.gc.reprocess_uows



__author__ = 'Bohdan Mushkevych'

import json
import httplib

from werkzeug.wrappers import Response
from werkzeug.utils import redirect

from synergy.mx.gc_action_handler import GcActionHandler
from synergy.mx.freerun_action_handler import FreerunActionHandler
from synergy.mx.managed_action_handler import ManagedActionHandler
from synergy.mx.scheduler_entries import SchedulerEntries
from synergy.mx.dashboard_handler import DashboardHandler
from synergy.mx.utils import render_template, expose
from synergy.mx.tree_node_details import TreeNodeDetails
from synergy.mx.tree_details import TreeDetails


@expose('/entries/managed/')
def scheduler_managed_entries(request, **values):
    details = SchedulerEntries(request, **values)
    return render_template('scheduler_managed_entries.html', details=details)


@expose('/entries/freerun/')
def scheduler_freerun_entries(request, **values):
    details = SchedulerEntries(request, **values)
    return render_template('scheduler_freerun_entries.html', details=details)


@expose('/open_schedulable_form/')
def open_schedulable_form(request, **values):
    if 'is_new_entry' in request.args and request.args['is_new_entry'] in ('True', 'true', '1'):
        handler = None
    else:
        handler = FreerunActionHandler(request, **values)
    return render_template('schedulable_form.html', handler=handler)

@expose('/')
def landing_page(request, **values):
    return redirect('/dashboard/managed/')


@expose('/dashboard/managed/')
def dashboard_managed(request, **values):
    details = DashboardHandler(request, **values)
    return render_template('dashboard_managed.html', details=details)


@expose('/dashboard/freeruns/')
def dashboard_freeruns(request, **values):
    details = DashboardHandler(request, **values)
    return render_template('dashboard_freeruns.html', details=details)


@expose('/details/tree_nodes/')
def details_tree_nodes(request, **values):
    details = TreeNodeDetails(request, **values)
    return Response(response=json.dumps(details.details), mimetype='application/json')


@expose('/details/tree/')
def details_tree(request, **values):
    details = TreeDetails(request, **values)
    return Response(response=json.dumps(details.tree_details), mimetype='application/json')


@expose('/details/trees/')
def details_trees(request, **values):
    details = TreeDetails(request, **values)
    return Response(response=json.dumps(details.mx_page_entries), mimetype='application/json')


@expose('/action/update_freerun_entry/')
def action_update_freerun_entry(request, **values):
    handler = FreerunActionHandler(request, **values)
    handler.action_update_entry()
    return Response(status=httplib.NO_CONTENT)


@expose('/action/reprocess/')
def action_reprocess(request, **values):
    handler = ManagedActionHandler(request, **values)
    handler.action_reprocess()
    return Response(status=httplib.NO_CONTENT)


@expose('/action/skip/')
def action_skip(request, **values):
    handler = ManagedActionHandler(request, **values)
    handler.action_skip()
    return Response(status=httplib.NO_CONTENT)


@expose('/action/cancel_uow/')
def action_cancel_uow(request, **values):
    handler = FreerunActionHandler(request, **values)
    handler.action_cancel_uow()
    return Response(status=httplib.NO_CONTENT)


@expose('/action/get_uow/')
def action_get_uow(request, **values):
    handler = get_action_handler(request, **values)
    return Response(response=json.dumps(handler.action_get_uow()), mimetype='application/json')


@expose('/action/get_log/')
def action_get_log(request, **values):
    handler = get_action_handler(request, **values)
    return Response(response=json.dumps(handler.action_get_log()), mimetype='application/json')


@expose('/action/change_interval/')
def action_change_interval(request, **values):
    handler = get_action_handler(request, **values)
    handler.action_change_interval()
    return Response(status=httplib.NO_CONTENT)


@expose('/action/trigger_now/')
def action_trigger_now(request, **values):
    handler = get_action_handler(request, **values)
    handler.action_trigger_now()
    return Response(status=httplib.NO_CONTENT)


@expose('/action/deactivate_trigger/')
def action_deactivate_trigger(request, **values):
    handler = get_action_handler(request, **values)
    handler.action_deactivate_trigger()
    return Response(status=httplib.NO_CONTENT)


@expose('/action/activate_trigger/')
def action_activate_trigger(request, **values):
    handler = get_action_handler(request, **values)
    handler.action_activate_trigger()
    return Response(status=httplib.NO_CONTENT)


@expose('/gc/flush_all/')
def gc_flush_all(request, **values):
    handler = GcActionHandler(request, **values)
    handler.action_flush_all()
    return Response(status=httplib.NO_CONTENT)


@expose('/gc/flush_one/')
def gc_flush_one(request, **values):
    handler = GcActionHandler(request, **values)
    handler.action_flush_one()
    return Response(status=httplib.NO_CONTENT)


@expose('/gc/refresh/')
def gc_refresh(request, **values):
    handler = GcActionHandler(request, **values)
    handler.action_refresh()
    return Response(status=httplib.NO_CONTENT)


def get_action_handler(request, **values):
    if 'is_freerun' in request.args and request.args['is_freerun'] in ('True', 'true', '1'):
        handler = FreerunActionHandler(request, **values)
    else:
        handler = ManagedActionHandler(request, **values)
    return handler


@expose('/object_viewer/')
def object_viewer(request, **values):
    return render_template('object_viewer.html')


@expose('/mx_page_tiles/')
def mx_page_tiles(request, **values):
    details = TreeDetails(request, **values)
    return render_template('mx_page_tiles.html', details=details.mx_page_entries)


# referenced from mx.synergy_mx.py module
def not_found(request, **values):
    return render_template('not_found.html')



__author__ = 'Bohdan Mushkevych'

import sys
import logging
import logging.handlers

from synergy.conf import settings
from synergy.conf import context
from synergy.db.model.daemon_process_entry import DaemonProcessEntry
from synergy.db.model.freerun_process_entry import FreerunProcessEntry
from synergy.db.model.managed_process_entry import ManagedProcessEntry


class Logger(object):
    """ Logger presents wrapper around standard API enriched with formaters and roto handlers """

    def __init__(self, file_name, log_tag, append_to_console, redirect_stdstream):
        """
        :param file_name: path+name of the output file
        :param log_tag: tag that is printed ahead of every logged message
        :param append_to_console: True if messages should be printed to the terminal console
        :param redirect_stdstream: True if stdout and stderr should be redirected to this Logger instance
        """
        self.logger = logging.getLogger(log_tag)

        if append_to_console:
            # ATTENTION: while running as stand-alone process, stdout and stderr must be muted and redirected to file
            # otherwise the their pipes get overfilled, and process halts
            stream_handler = logging.StreamHandler()
            stream_formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
            stream_handler.setFormatter(stream_formatter)
            self.logger.addHandler(stream_handler)

        if redirect_stdstream:
            # While under_test, tools as xml_unittest_runner are doing complex sys.stdXXX reassignments
            sys.stdout = self
            sys.stderr = self

        if settings.settings['debug']:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

        # ROTO FILE HANDLER:
        roto_file_handler = logging.handlers.RotatingFileHandler(file_name, maxBytes=2097152, backupCount=10)
        roto_file_formatter = logging.Formatter(fmt='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                                                datefmt='%Y-%m-%d %H:%M:%S')
        roto_file_handler.setFormatter(roto_file_formatter)
        self.logger.addHandler(roto_file_handler)

    def get_logger(self):
        return self.logger

    def write(self, msg, level=logging.INFO):
        """ method implements stream write interface, allowing to redirect stdout to logger """
        if msg is not None and len(msg.strip()) > 0:
            self.logger.log(level, msg)

    def flush(self):
        """ method implements stream flush interface, allowing to redirect stdout to logger """
        for handler in self.logger.handlers:
            handler.flush()

    def isatty(self):
        """ is the sys.stdout attached to the terminal?
        python -c "import sys; print(sys.stdout.isatty())" (should write True)
        python -c "import sys; print(sys.stdout.isatty())" | grep . (should write False).
        :return: False, indicating that the output is pipped or redirected
        """
        return False


# holds Logger instance per process name (and optional suffix)
logger_pool = dict()


def get_logger(process_name,
               append_to_console=settings.settings['under_test'],
               redirect_stdstream=not settings.settings['under_test']):
    """ method returns initiated logger"""
    if process_name not in logger_pool:
        file_name = get_log_filename(process_name)
        log_tag = get_log_tag(process_name)
        logger_pool[process_name] = Logger(file_name, log_tag,
                                           append_to_console=append_to_console,
                                           redirect_stdstream=redirect_stdstream)
    return logger_pool[process_name].get_logger()


def get_log_filename(process_name):
    """method returns path for the Log filename"""
    return settings.settings['log_directory'] + context.process_context[process_name].log_filename


def get_log_tag(process_name):
    """method returns tag that all messages will be preceded with"""
    process_obj = context.process_context[process_name]
    if isinstance(process_obj, FreerunProcessEntry):
        return str(process_obj.token)
    elif isinstance(process_obj, ManagedProcessEntry):
        return str(process_obj.token) + str(process_obj.time_qualifier)
    elif isinstance(process_obj, DaemonProcessEntry):
        return str(process_obj.token)
    else:
        raise ValueError('Unknown process type: {0}'.format(process_obj.__class__.__name__))


if __name__ == '__main__':
    process_name = 'TestAggregator'
    logger = get_logger(process_name)
    logger.info('test_message')
    print('regular print message')
    sys.stdout.flush()



import json
import uuid
from collections import deque
from threading import Lock

import amqp
import sys

from synergy.conf import settings, context
from synergy.system.decorator import thread_safe


class SynergyAware(object):
    def __init__(self, name):
        self.name = name
        if name in context.process_context:
            the_context = context.process_context
        elif name in context.mq_queue_context:
            the_context = context.mq_queue_context
        else:
            raise ValueError('Unknown name {0}. Unable to retrieve amqp settings.'.format(name))

        self.routing_key = the_context[name].mq_routing_key
        self.exchange = the_context[name].mq_exchange
        self.queue = the_context[name].mq_queue


class Connection(object):
    def __init__(self,
                 host=settings.settings['mq_host'],
                 user_id=settings.settings['mq_user_id'],
                 password=settings.settings['mq_password'],
                 vhost=settings.settings['mq_vhost'],
                 port=settings.settings['mq_port']):
        self.db_host = host
        self.user_id = user_id
        self.password = password
        self.vhost = vhost
        self.port = port
        self.connection = None

        self.connect()

    def __del__(self):
        self.close()

    def connect(self):
        self.connection = amqp.Connection(
            host='{0}:{1}'.format(self.db_host, self.port),
            userid=self.user_id,
            password=self.password,
            virtual_host=self.vhost
        )

    def close(self):
        if hasattr(self, 'connection') and self.connection:
            self.connection.close()


class Consumer(SynergyAware):
    def __init__(self,
                 name,
                 durable=settings.settings['mq_durable'],
                 exclusive=settings.settings['mq_exclusive'],
                 auto_delete=settings.settings['mq_auto_delete'],
                 connection=None):
        super(Consumer, self).__init__(name)
        self.callback = None
        self.is_running = True

        self.durable = durable
        self.exclusive = exclusive
        self.auto_delete = auto_delete
        self.connection = connection or Connection()
        self.channel = self.connection.connection.channel()

        self.channel.queue_declare(
            queue=self.queue,
            durable=self.durable,
            exclusive=self.exclusive,
            auto_delete=self.auto_delete
        )
        self.channel.exchange_declare(
            exchange=self.exchange,
            type='direct',
            durable=self.durable,
            auto_delete=self.auto_delete
        )
        self.channel.queue_bind(
            queue=self.queue,
            exchange=self.exchange,
            routing_key=self.routing_key
        )
        self.channel.basic_consume(
            queue=self.queue,
            no_ack=settings.settings['mq_no_ack'],
            callback=self.dispatch,
            consumer_tag=str(uuid.uuid4())
        )

    def __del__(self):
        self.close()

    def close(self):
        self.is_running = False

        if hasattr(self, 'channel') and self.channel:
            self.channel.close()

        if hasattr(self, 'connection') and self.connection:
            self.connection.close()

    def wait(self, timeout=None):
        while self.is_running:
            channel_id, method_sig, args, content = \
                self.connection.connection._wait_multiple(channels={self.channel.channel_id: self.channel},
                                                          allowed_methods=None,
                                                          timeout=timeout)
            self.channel.dispatch_method(method_sig, args, content)

    def dispatch(self, message):
        decoded = json.loads(message.body)
        message.body = decoded['data']
        if self.callback is not None:
            self.callback(message)

    def acknowledge(self, tag):
        if settings.settings['mq_no_ack'] is False:
            self.channel.basic_ack(delivery_tag=tag)

    def reject(self, tag):
        if settings.settings['mq_no_ack'] is False:
            self.channel.basic_reject(delivery_tag=tag, requeue=True)

    def cancel(self, tag):
        if settings.settings['mq_no_ack'] is False:
            self.channel.basic_reject(delivery_tag=tag, requeue=False)

    def register(self, callback):
        self.callback = callback

    def unregister(self):
        self.callback = None


class Publisher(SynergyAware):
    def __init__(self,
                 name,
                 connection=None,
                 delivery_mode=settings.settings['mq_delivery_mode'],
                 parent_pool=None):
        super(Publisher, self).__init__(name)
        self.connection = connection or Connection()
        self.channel = self.connection.connection.channel()
        self.delivery_mode = delivery_mode
        self.parent_pool = parent_pool

    def __del__(self):
        self.release()

    def publish(self, message_data):
        encoded = json.dumps({'data': message_data})
        message = amqp.Message(encoded)
        message.properties['delivery_mode'] = self.delivery_mode
        self.channel.basic_publish(
            message,
            exchange=self.exchange,
            routing_key=self.routing_key
        )
        return message

    def release(self):
        if hasattr(self, 'parent_pool') and self.parent_pool:
            self.parent_pool.put(self)
        else:
            self.close()

    def close(self):
        if hasattr(self, 'channel') and self.channel:
            self.channel.close()

        if hasattr(self, 'connection') and self.connection:
            self.connection.close()

        if hasattr(self, 'parent_pool') and self.parent_pool:
            del self.parent_pool


class _Pool(object):
    def __init__(self, logger, name):
        self.publishers = deque()
        self.name = name
        self.logger = logger
        self.lock = Lock()

    def __del__(self):
        self.close()

    @thread_safe
    def get(self):
        """ :return valid :mq::flopsy::Publisher instance """
        if len(self.publishers) == 0:
            return Publisher(name=self.name, parent_pool=self)
        else:
            return self.publishers.pop()

    @thread_safe
    def put(self, publisher):
        self.publishers.append(publisher)

    @thread_safe
    def close(self, suppress_logging=False):
        """ purges all connections. method closes ampq connection (disconnects) """
        for publisher in self.publishers:
            try:
                publisher.close()
            except Exception as e:
                self.logger.error('Exception on closing Flopsy Publisher {0}: {1}'.format(self.name, e),
                                  exc_info=not suppress_logging)
        self.publishers.clear()


class PublishersPool(object):
    def __init__(self, logger):
        self.pools = dict()
        self.logger = logger

    def __del__(self):
        self.close(suppress_logging=True)

    def get(self, name):
        """ creates connection to the MQ with process-specific settings
        :return :mq::flopsy::Publisher instance"""
        if name not in self.pools:
            self.pools[name] = _Pool(logger=self.logger, name=name)
        return self.pools[name].get()

    def put(self, publisher):
        """ releases the Publisher instance for reuse"""
        if publisher.name not in self.pools:
            self.pools[publisher.name] = _Pool(logger=self.logger, name=publisher.name)
        self.pools[publisher.name].put(publisher)

    def reset_all(self, suppress_logging=False):
        """ iterates thru the list of established connections and resets them by disconnecting and reconnecting """
        pool_names = self.pools.keys()
        for name in pool_names:
            self.reset(name, suppress_logging)

    def reset(self, name, suppress_logging=False):
        """ resets established connection by disconnecting and reconnecting """
        self._close(name, suppress_logging)
        self.get(name)
        self.logger.info('Reset Flopsy Pool for {0}'.format(name))

    def _close(self, name, suppress_logging):
        """ closes one particular pool and all its amqp amqp connections """
        try:
            pool_names = self.pools.keys()
            if name in pool_names:
                self.pools[name].close()
                del self.pools[name]
        except Exception as e:
            self.logger.error('Exception on closing Flopsy Pool for {0}: {1}'.format(name, e),
                              exc_info=not suppress_logging)

    def close(self, suppress_logging=False):
        """ iterates thru all publisher pools and closes them """
        pool_names = self.pools.keys()
        for name in pool_names:
            self._close(name, suppress_logging)


def purge_mq_queue(mq_queue_name):
    """ function purges
    :param mq_queue_name: <string> name of the message queue
    :return: number of purged messages
    """
    conn = None
    chan = None
    try:
        conn = Connection()
        chan = conn.connection.channel()
        n = chan.queue_purge(mq_queue_name)
        sys.stdout.write('Purged {0} messages from {1} queue\n'.format(n, mq_queue_name))
        return n
    except Exception as e:
        sys.stderr.write('Unable to purge {0} due to {1}\n'.format(mq_queue_name, e))
    finally:
        if chan is not None:
            chan.close()
        if conn is not None:
            conn.close()



__author__ = 'Bohdan Mushkevych'

from datetime import datetime
from logging import ERROR, WARNING, INFO

from synergy.conf import context
from synergy.db.error import DuplicateKeyError
from synergy.db.model import unit_of_work
from synergy.db.model.unit_of_work import UnitOfWork
from synergy.db.model.synergy_mq_transmission import SynergyMqTransmission
from synergy.db.model.freerun_process_entry import FreerunProcessEntry, MAX_NUMBER_OF_LOG_ENTRIES
from synergy.db.dao.unit_of_work_dao import UnitOfWorkDao
from synergy.db.dao.freerun_process_dao import FreerunProcessDao
from synergy.system import time_helper
from synergy.system.time_qualifier import QUALIFIER_REAL_TIME
from synergy.system.decorator import with_reconnect
from synergy.scheduler.scheduler_constants import STATE_MACHINE_FREERUN
from synergy.mq.flopsy import PublishersPool


class StateMachineFreerun(object):
    """ State Machine to handle freerun jobs/triggers """

    def __init__(self, logger, name=STATE_MACHINE_FREERUN):
        self.name = name
        self.logger = logger
        self.publishers = PublishersPool(self.logger)
        self.uow_dao = UnitOfWorkDao(self.logger)
        self.sfe_dao = FreerunProcessDao(self.logger)

    def __del__(self):
        try:
            self.logger.info('Closing Flopsy Publishers Pool...')
            self.publishers.close()
        except Exception as e:
            self.logger.error('Exception caught while closing Flopsy Publishers Pool: {0}'.format(e))

    @with_reconnect
    def _log_message(self, level, freerun_entry, msg):
        """ method performs logging into log file and the freerun_entry """
        self.logger.log(level, msg)

        assert isinstance(freerun_entry, FreerunProcessEntry)
        log = freerun_entry.log
        if len(log) > MAX_NUMBER_OF_LOG_ENTRIES:
            del log[-1]
        log.insert(0, msg)
        self.sfe_dao.update(freerun_entry)

    @with_reconnect
    def _insert_uow(self, freerun_entry):
        """ creates unit_of_work and inserts it into the DB
            :raise DuplicateKeyError: if unit_of_work with given parameters already exists """
        current_timeperiod = time_helper.actual_timeperiod(QUALIFIER_REAL_TIME)
        process_entry = context.process_context[freerun_entry.process_name]

        uow = UnitOfWork()
        uow.process_name = freerun_entry.schedulable_name
        uow.timeperiod = current_timeperiod
        uow.start_id = 0
        uow.end_id = 0
        uow.start_timeperiod = current_timeperiod
        uow.end_timeperiod = current_timeperiod
        uow.created_at = datetime.utcnow()
        uow.submitted_at = datetime.utcnow()
        uow.source = process_entry.source if hasattr(process_entry, 'source') else None
        uow.sink = process_entry.sink if hasattr(process_entry, 'sink') else None
        uow.state = unit_of_work.STATE_REQUESTED
        uow.unit_of_work_type = unit_of_work.TYPE_FREERUN
        uow.number_of_retries = 0
        uow.arguments = freerun_entry.arguments
        uow.db_id = self.uow_dao.insert(uow)

        msg = 'Created: UOW {0} for {1}@{2}.'\
              .format(uow.db_id, freerun_entry.schedulable_name, current_timeperiod)
        self._log_message(INFO, freerun_entry, msg)
        return uow

    def _publish_uow(self, freerun_entry, uow):
        mq_request = SynergyMqTransmission(process_name=freerun_entry.process_name,
                                           entry_name=freerun_entry.entry_name,
                                           unit_of_work_id=uow.db_id)

        publisher = self.publishers.get(freerun_entry.process_name)
        publisher.publish(mq_request.document)
        publisher.release()

        msg = 'Published: UOW {0} for {1}.'.format(uow.db_id, freerun_entry.schedulable_name)
        self._log_message(INFO, freerun_entry, msg)

    def insert_and_publish_uow(self, freerun_entry):
        try:
            uow = self._insert_uow(freerun_entry)
        except DuplicateKeyError as e:
            msg = 'Duplication of UOW found for {0}. Error msg: {1}'.format(freerun_entry.schedulable_name, e)
            self._log_message(WARNING, freerun_entry, msg)
            uow = self.uow_dao.recover_from_duplicatekeyerror(e)

        if uow is not None:
            # publish the created/caught up unit_of_work
            self._publish_uow(freerun_entry, uow)
            freerun_entry.related_unit_of_work = uow.db_id
            self.sfe_dao.update(freerun_entry)
        else:
            msg = 'PERSISTENT TIER ERROR! Unable to locate UOW for {0}'\
                  .format(freerun_entry.schedulable_name)
            self._log_message(WARNING, freerun_entry, msg)

    def _process_state_embryo(self, freerun_entry):
        """ method creates unit_of_work and associates it with the FreerunProcessEntry """
        self.insert_and_publish_uow(freerun_entry)

    def _process_state_in_progress(self, freerun_entry, uow):
        """ method that takes care of processing unit_of_work records in STATE_REQUESTED or STATE_IN_PROGRESS states"""
        self._publish_uow(freerun_entry, uow)

    def _process_terminal_state(self, freerun_entry, uow):
        """ method that takes care of processing unit_of_work records in
            STATE_PROCESSED, STATE_NOOP, STATE_INVALID, STATE_CANCELED states"""
        msg = 'UOW for {0} found in state {1}.'.format(freerun_entry.schedulable_name, uow.state)
        self._log_message(INFO, freerun_entry, msg)
        self.insert_and_publish_uow(freerun_entry)

    def manage_schedulable(self, freerun_entry):
        """ method main duty - is to _avoid_ publishing another unit_of_work, if previous was not yet processed
        In case the Scheduler sees that the unit_of_work is pending it will fire another WorkerMqRequest """

        assert isinstance(freerun_entry, FreerunProcessEntry)
        if freerun_entry.related_unit_of_work is None:
            uow = None
        else:
            uow = self.uow_dao.get_one(freerun_entry.related_unit_of_work)

        try:
            if uow is None:
                self._process_state_embryo(freerun_entry)

            elif uow.is_requested or uow.is_in_progress:
                self._process_state_in_progress(freerun_entry, uow)

            elif uow.is_finished or uow.is_invalid:
                self._process_terminal_state(freerun_entry, uow)

            else:
                msg = 'Unknown state {0} of the UOW {1}'.format(uow.state, uow.db_id)
                self._log_message(ERROR, freerun_entry, msg)

        except LookupError as e:
            msg = 'Lookup issue for schedulable: {0} in timeperiod {1}, because of: {2}'\
                  .format(freerun_entry.db_id, uow.timeperiod, e)
            self._log_message(WARNING, freerun_entry, msg)

    def cancel_uow(self, freerun_entry):
        uow_id = freerun_entry.related_unit_of_work
        if uow_id is None:
            msg = 'cancel_uow: no related UOW for {0}'.format(freerun_entry.schedulable_name)
        else:
            uow = self.uow_dao.get_one(uow_id)
            uow.state = unit_of_work.STATE_CANCELED
            self.uow_dao.update(uow)
            msg = 'cancel_uow: canceled UOW {0} for {1}'.format(uow_id, freerun_entry.schedulable_name)
        self._log_message(INFO, freerun_entry, msg)



__author__ = 'Bohdan Mushkevych'

from synergy.mx.base_request_handler import BaseRequestHandler, valid_action_request


class AbstractActionHandler(BaseRequestHandler):
    def __init__(self, request, **values):
        super(AbstractActionHandler, self).__init__(request, **values)

    @property
    def thread_handler(self):
        raise NotImplementedError('property thread_handler must be implemented by {0}'.format(self.__class__.__name__))

    @property
    def process_entry(self):
        raise NotImplementedError('property process_entry must be implemented by {0}'.format(self.__class__.__name__))

    def action_get_uow(self):
        raise NotImplementedError('method action_get_uow must be implemented by {0}'.format(self.__class__.__name__))

    def action_get_log(self):
        raise NotImplementedError('method action_get_log must be implemented by {0}'.format(self.__class__.__name__))

    @valid_action_request
    def action_change_interval(self):
        resp = dict()
        new_interval = self.request_arguments['interval']
        if new_interval is not None:
            self.thread_handler.change_interval(new_interval)
            msg = 'changed interval for {0} to {1}'.format(self.thread_handler.key, new_interval)
            self.logger.info('MX: {0}'.format(msg))
            resp['status'] = msg

        return resp

    @valid_action_request
    def action_trigger_now(self):
        self.thread_handler.trigger()
        self.logger.info('MX: triggered thread handler {0}'.format(self.thread_handler.key))
        return self.reply_ok()

    @valid_action_request
    def action_activate_trigger(self):
        self.thread_handler.activate()
        self.logger.info('MX: activated thread handler {0}'.format(self.thread_handler.key))
        return self.reply_ok()

    @valid_action_request
    def action_deactivate_trigger(self):
        self.thread_handler.deactivate()
        self.logger.info('MX: deactivated thread handler {0}'.format(self.thread_handler.key))
        return self.reply_ok()



__author__ = 'Bohdan Mushkevych'

from threading import RLock

from bson import ObjectId
from synergy.db.manager import ds_manager
from synergy.db.model import job
from synergy.db.model.job import Job
from synergy.system.decorator import thread_safe
from synergy.system.time_qualifier import *
from synergy.scheduler.scheduler_constants import COLLECTION_JOB_HOURLY, COLLECTION_JOB_DAILY, \
    COLLECTION_JOB_MONTHLY, COLLECTION_JOB_YEARLY
from synergy.conf import context


QUERY_GET_LIKE_TIMEPERIOD = lambda timeperiod, include_running, include_processed, include_noop, include_failed: {
    job.TIMEPERIOD: {'$gte': timeperiod},
    job.STATE: {'$in': [job.STATE_PROCESSED if include_processed else None,
                        job.STATE_IN_PROGRESS if include_running else None,
                        job.STATE_EMBRYO if include_running else None,
                        job.STATE_FINAL_RUN if include_running else None,
                        job.STATE_SKIPPED if include_failed else None,
                        job.STATE_NOOP if include_noop else None]}
}


class JobDao(object):
    """ Thread-safe Data Access Object from job_XXX collection
    above, XXX could stand for hourly, daily, monthly, yearly """
    def __init__(self, logger):
        super(JobDao, self).__init__()
        self.logger = logger
        self.lock = RLock()
        self.ds = ds_manager.ds_factory(logger)

    @thread_safe
    def _get_job_collection(self, process_name):
        """jobs are stored in 4 collections: hourly, daily, monthly and yearly;
        method looks for the proper job_collection base on process TIME_QUALIFIER"""
        qualifier = context.process_context[process_name].time_qualifier

        if qualifier == QUALIFIER_HOURLY:
            collection = self.ds.connection(COLLECTION_JOB_HOURLY)
        elif qualifier == QUALIFIER_DAILY:
            collection = self.ds.connection(COLLECTION_JOB_DAILY)
        elif qualifier == QUALIFIER_MONTHLY:
            collection = self.ds.connection(COLLECTION_JOB_MONTHLY)
        elif qualifier == QUALIFIER_YEARLY:
            collection = self.ds.connection(COLLECTION_JOB_YEARLY)
        else:
            raise ValueError('Unknown time qualifier: {0} for {1}'.format(qualifier, process_name))
        return collection

    @thread_safe
    def get_one(self, key, timeperiod):
        """ method finds job record and returns it to the caller"""
        collection = self._get_job_collection(key)
        document = collection.find_one({job.PROCESS_NAME: key, job.TIMEPERIOD: timeperiod})

        if document is None:
            raise LookupError('MongoDB has no job record in collection {0} for {1}@{2}'
                              .format(collection, key, timeperiod))
        return Job.from_json(document)

    @thread_safe
    def get_all(self, collection_name, since=None):
        """ method returns all job records from a particular collection that are older than <since> """
        if since is None:
            query = {}
        else:
            query = {job.TIMEPERIOD: {'$gte': since}}
        collection = self.ds.connection(collection_name)

        cursor = collection.find(query)
        if cursor.count() == 0:
            raise LookupError('MongoDB has no job records in collection {0} since {1}'.format(collection_name, since))
        return [Job.from_json(document) for document in cursor]

    @thread_safe
    def run_query(self, collection_name, query):
        """ method runs query on a specified collection and return a list of filtered Job records """
        cursor = self.ds.filter(collection_name, query)
        return [Job.from_json(document) for document in cursor]

    @thread_safe
    def update(self, instance):
        assert isinstance(instance, Job)
        collection = self._get_job_collection(instance.process_name)
        document = instance.document
        if instance.db_id:
            document['_id'] = ObjectId(instance.db_id)
        instance.db_id = collection.save(document, safe=True)
        return instance.db_id



__author__ = 'Raymond Hettinger'
# http://stackoverflow.com/a/9997519/3171310

import collections


class ImmutableDict(collections.Mapping):
    def __init__(self, somedict):
        self._dict = dict(somedict)   # make a copy
        self._hash = None

    def __getitem__(self, key):
        return self._dict[key]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(frozenset(self._dict.items()))
        return self._hash

    def __eq__(self, other):
        return self._dict == other._dict



__author__ = 'Bohdan Mushkevych'

from synergy.db.dao.base_dao import BaseDao
from synergy.db.model.freerun_process_entry import FreerunProcessEntry, PROCESS_NAME, ENTRY_NAME
from synergy.scheduler.scheduler_constants import COLLECTION_FREERUN_PROCESS


class FreerunProcessDao(BaseDao):
    """ Thread-safe Data Access Object for freerun_process table/collection """

    def __init__(self, logger):
        super(FreerunProcessDao, self).__init__(logger=logger,
                                                model_class=FreerunProcessEntry,
                                                primary_key=[PROCESS_NAME, ENTRY_NAME],
                                                collection_name=COLLECTION_FREERUN_PROCESS)



__author__ = 'Bohdan Mushkevych'

from bson import ObjectId
from threading import RLock

from synergy.db.manager import ds_manager
from synergy.system.decorator import thread_safe


class BaseDao(object):
    """ Thread-safe base Data Access Object """

    def __init__(self, logger, model_class, primary_key, collection_name):
        super(BaseDao, self).__init__()
        self.logger = logger
        self.model_klass = model_class
        self.primary_key = primary_key
        self.collection_name = collection_name

        self.lock = RLock()
        self.ds = ds_manager.ds_factory(logger)

    def _tuple_to_query(self, key_tuple):
        if isinstance(key_tuple, str):
            key_tuple = [key_tuple]
        assert len(key_tuple) == len(self.primary_key)

        query = dict()
        for k, v in zip(self.primary_key, key_tuple):
            query[k] = v
        return query

    @thread_safe
    def get_one(self, key):
        """ method finds single record base on the given primary key and returns it to the caller"""
        query = self._tuple_to_query(key)
        collection = self.ds.connection(self.collection_name)

        document = collection.find_one(query)
        if document is None:
            raise LookupError('{0} with key {1} was not found'.format(self.model_klass.__name__, query))
        return self.model_klass.from_json(document)

    @thread_safe
    def run_query(self, query):
        """ method runs query on a specified collection and return a list of filtered Model records """
        collection = self.ds.connection(self.collection_name)

        cursor = collection.find(query)
        if cursor.count() == 0:
            raise LookupError('Collection {0} has no {1} records'
                              .format(self.collection_name, self.model_klass.__name__))
        return [self.model_klass.from_json(entry) for entry in cursor]

    @thread_safe
    def get_all(self):
        return self.run_query({})

    @thread_safe
    def update(self, instance, is_safe=True):
        """ this is upsert method: inserts or updates the DB representation of the model instance """
        assert isinstance(instance, self.model_klass)
        collection = self.ds.connection(self.collection_name)
        document = instance.document
        if instance.db_id:
            document['_id'] = ObjectId(instance.db_id)
        instance.db_id = collection.save(document, safe=is_safe)
        return instance.db_id

    @thread_safe
    def remove(self, key):
        query = self._tuple_to_query(key)
        collection = self.ds.connection(self.collection_name)
        return collection.remove(query, safe=True)



__author__ = 'Bohdan Mushkevych'

from threading import Thread
from werkzeug.wrappers import Request
from werkzeug.wsgi import ClosingIterator, SharedDataMiddleware
from werkzeug.exceptions import HTTPException, NotFound
from werkzeug.serving import run_simple
from synergy.conf import settings
from synergy.system.data_logging import get_logger
from synergy.scheduler.scheduler_constants import PROCESS_MX

from synergy.mx.utils import STATIC_PATH, local, local_manager, url_map, jinja_env
from synergy.mx import views


class MX(object):
    """ MX stands for Management Extension and represents HTTP server serving UI front-end for Synergy Scheduler """
    def __init__(self, mbean):
        local.application = self
        self.mx_thread = None
        self.mbean = mbean
        jinja_env.globals['mbean'] = mbean

        self.dispatch = SharedDataMiddleware(self.dispatch, {
            '/static': STATIC_PATH
        })

        # during the get_logger call a 'werkzeug' logger will be created
        # later, werkzeug._internal.py -> _log() will assign the logger to global _logger variable
        self.logger = get_logger(PROCESS_MX)

    def dispatch(self, environ, start_response):
        local.application = self
        request = Request(environ)
        local.url_adapter = adapter = url_map.bind_to_environ(environ)
        local.request = request

        try:
            endpoint, values = adapter.match()
            handler = getattr(views, endpoint)
            response = handler(request, **values)
        except NotFound:
            response = views.not_found(request)
            response.status_code = 404
        except HTTPException as e:
            response = e
        return ClosingIterator(response(environ, start_response),
                               [local_manager.cleanup])

    def __call__(self, environ, start_response):
        return self.dispatch(environ, start_response)

    def start(self, hostname=None, port=None):
        """ Spawns a new HTTP server, residing on defined hostname and port
        :param hostname: the default hostname the server should listen on.
        :param port: the default port of the server.
        """
        if hostname is None:
            hostname = settings.settings['mx_host']
        if port is None:
            port = settings.settings['mx_port']

        reloader = False        # use_reloader: the default setting for the reloader.
        debugger = False        #
        evalex = True           # should the exception evaluation feature be enabled?
        threaded = False        # True if each request in handled in a separate thread
        processes = 1           # if greater than 1 then handle each request in a new process
        reloader_interval = 1   # the interval for the reloader in seconds.
        static_files = None     # static_files: optional dict of static files.
        extra_files = None      # extra_files: optional list of extra files to track for reloading.
        ssl_context = None      # ssl_context: optional SSL context for running server in HTTPS mode.

        self.mx_thread = Thread(target=run_simple(hostname=hostname,
                                                  port=port,
                                                  application=self,
                                                  use_debugger=debugger,
                                                  use_evalex=evalex,
                                                  extra_files=extra_files,
                                                  use_reloader=reloader,
                                                  reloader_interval=reloader_interval,
                                                  threaded=threaded,
                                                  processes=processes,
                                                  static_files=static_files,
                                                  ssl_context=ssl_context))
        self.mx_thread.daemon = True
        self.mx_thread.start()

    def stop(self):
        """ method stops currently running HTTP server, if any
            http://flask.pocoo.org/snippets/67/ """
        func = jinja_env.get('werkzeug.server.shutdown')
        if func is None:
            raise RuntimeError('MX Error: no Shutdown Function registered for the Werkzeug Server')
        func()

if __name__ == '__main__':
    from synergy.scheduler.scheduler_constants import PROCESS_SCHEDULER
    from synergy.scheduler.synergy_scheduler import Scheduler

    scheduler = Scheduler(PROCESS_SCHEDULER)
    app = MX(scheduler)
    app.start()



"""
@author: Bohdan Mushkevych
@author: Brian Curtin
http://code.activestate.com/lists/python-ideas/8982/
"""
import numbers
import threading
from datetime import datetime, timedelta


class RepeatTimer(threading.Thread):
    """ This class triggers every number of seconds """
    def __init__(self, interval, call_back, args=None, kwargs=None):
        if not kwargs: kwargs = {}
        if not args: args = []
        
        threading.Thread.__init__(self)
        assert isinstance(interval, numbers.Number)
        # interval_current shows number of seconds in currently triggered <tick>
        self.interval_current = interval
        # interval_new shows number of seconds for next <tick>
        self.interval_new = interval
        self.call_back = call_back
        self.args = args
        self.kwargs = kwargs
        self.event = threading.Event()
        self.event.set()
        self.activation_dt = None
        self.__timer = None

    def run(self):
        while self.event.is_set():
            self.activation_dt = datetime.utcnow()
            self.__timer = threading.Timer(self.interval_new,
                                           self.call_back,
                                           self.args,
                                           self.kwargs)
            self.interval_current = self.interval_new
            self.__timer.start()
            self.__timer.join()

    def cancel(self):
        """ stops the timer. call_back function is not called """
        self.event.clear()
        if self.__timer is not None:
            self.__timer.cancel()

    def trigger(self):
        """ calls the call_back function. interrupts the timer to start a new countdown """
        self.call_back(*self.args, **self.kwargs)
        if self.__timer is not None:
            self.__timer.cancel()

    def change_interval(self, value):
        """ :param value: <tick> interval in seconds
            current countdown is not interrupted
            new interval will be applied after the trigger execution """
        self.interval_new = value

    def next_run_in(self, utc_now=None):
        """ :param utc_now: optional parameter to be used by Unit Tests as a definition of "now"
            :return: timedelta instance presenting amount of time before the trigger is triggered next time
         or None if the RepeatTimer instance is not running """
        if utc_now is None:
            utc_now = datetime.utcnow()

        if self.is_alive():
            next_run = timedelta(seconds=self.interval_current) + self.activation_dt
            return next_run - utc_now
        else:
            return None



__author__ = 'Bohdan Mushkevych'

import time

from psutil import TimeoutExpired
from synergy.db.model import unit_of_work
from synergy.workers.abstract_uow_aware_worker import AbstractUowAwareWorker

RETURN_CODE_CANCEL_UOW = 99998888
RETURN_CODE_NOOP_UOW = 77776666


class AbstractCliWorker(AbstractUowAwareWorker):
    """ Module contains common logic for Command Line Callers.
        It executes shell command and updates unit_of_work base on command's return code """

    def __init__(self, process_name):
        super(AbstractCliWorker, self).__init__(process_name)
        self.cli_process = None

    def __del__(self):
        super(AbstractCliWorker, self).__del__()

    def _start_process(self, start_timeperiod, end_timeperiod, arguments):
        raise NotImplementedError('method _start_process must be implemented by {0}'.format(self.__class__.__name__))

    def _poll_process(self):
        """ between death of a process and its actual termination lies poorly documented requirement -
            <purging process' io pipes and reading exit status>.
            this can be done either by os.wait() or process.wait()
            :return tuple (boolean: alive, int: return_code) """
        try:
            self.logger.warn(self.cli_process.stderr.read())
            self.logger.info(self.cli_process.stdout.read())
            return_code = self.cli_process.wait(timeout=0.01)
            if return_code is None:
                # process is already terminated
                self.logger.info('Process {0} is terminated'.format(self.process_name))
            else:
                # process is terminated; possibly by OS
                self.logger.info('Process {0} got terminated. Cleaning up'.format(self.process_name))
            self.cli_process = None
            return False, return_code
        except TimeoutExpired:
            # process is alive and OK
            return True, None
        except Exception:
            self.logger.error('Exception on polling: {0}'.format(self.process_name), exc_info=True)
            return False, 999

    def _process_uow(self, uow):
        self._start_process(uow.start_timeperiod, uow.end_timeperiod, uow.arguments)
        code = None
        alive = True
        while alive:
            alive, code = self._poll_process()
            time.sleep(0.1)

        self.logger.info('Command Line Command return code is {0}'.format(code))
        if code == 0:
            return 0, unit_of_work.STATE_PROCESSED
        elif code == RETURN_CODE_CANCEL_UOW:
            return 0, unit_of_work.STATE_CANCELED
        elif code == RETURN_CODE_NOOP_UOW:
            return 0, unit_of_work.STATE_NOOP
        else:
            raise UserWarning('Command Line Command return code is not 0 but {0}'.format(code))



from distutils.core import setup

setup(name='synergy_scheduler',
      version='1.13',
      description='Synergy Scheduler',
      author='Bohdan Mushkevych',
      author_email='mushkevych@gmail.com',
      url='https://github.com/mushkevych/scheduler',
      packages=['synergy', 'synergy.db', 'synergy.db.dao', 'synergy.db.manager', 'synergy.db.model',
                'synergy.mq', 'synergy.conf', 'synergy.mx', 'synergy.scheduler', 'synergy.supervisor',
                'synergy.system', 'synergy.workers'],
      package_data={'synergy.mx': ['static/images/*', 'static/fonts/*', 'static/js/*', 'static/css/*', 'templates/*'],
                    'synergy.mq': ['AUTHORS', 'LICENSE']},
      long_description='Synergy Scheduler shines in governing interdependent processes and their jobs. '
                       'Alternatively, it can work as a cron-like job trigger.',
      license='Modified BSD License',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Environment :: Web Environment',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: BSD License',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: JavaScript',
          'Topic :: Office/Business :: Scheduling',
          'Topic :: Utilities',
      ],
      requires=['werkzeug', 'jinja2', 'amqp', 'pymongo', 'psutil', 'fabric', 'setproctitle', 'synergy_odm', 'mock',
                'xmlrunner', 'pylint', 'six']
      )



__author__ = 'Bohdan Mushkevych'

from datetime import datetime, timedelta
from threading import Lock

from amqp import AMQPError

from synergy.conf import context
from synergy.mq.flopsy import PublishersPool
from synergy.mx.synergy_mx import MX
from synergy.db.manager import db_manager
from synergy.db.model.managed_process_entry import ManagedProcessEntry
from synergy.db.model.freerun_process_entry import FreerunProcessEntry
from synergy.db.dao.freerun_process_dao import FreerunProcessDao
from synergy.system import time_helper
from synergy.system.decorator import with_reconnect, thread_safe
from synergy.system.synergy_process import SynergyProcess
from synergy.scheduler.garbage_collector import GarbageCollector
from synergy.scheduler.status_bus_listener import StatusBusListener
from synergy.scheduler.scheduler_constants import *
from synergy.scheduler.timetable import Timetable
from synergy.scheduler.thread_handler import ThreadHandlerHeader, ManagedThreadHandler, FreerunThreadHandler


class Scheduler(SynergyProcess):
    """ Scheduler hosts:
        - timetable: container for job tress and state machines
        - GarbageCollector: recycles failed/stalled unit of works
        - freerun and managed thread handlers: logic to trigger job execution
        - StatusBarListener: MQ Listener of workers feedback
        - MX: HTTP server with management UI """

    def __init__(self, process_name):
        super(Scheduler, self).__init__(process_name)
        self.lock = Lock()
        self.logger.info('Starting {0}'.format(self.process_name))
        self.publishers = PublishersPool(self.logger)
        self.managed_handlers = dict()
        self.freerun_handlers = dict()
        self.timetable = Timetable(self.logger)
        self.freerun_process_dao = FreerunProcessDao(self.logger)

        self.gc = GarbageCollector(self)
        self.bus_listener = StatusBusListener(self)
        self.mx = MX(self)
        self.logger.info('Started {0}'.format(self.process_name))

    def __del__(self):
        self.mx.stop()
        self.bus_listener.stop()
        self.gc.stop()

        for key, handler in self.managed_handlers.items():
            handler.deactivate(update_persistent=False)
        self.managed_handlers.clear()

        for key, handler in self.freerun_handlers.items():
            handler.deactivate(update_persistent=False)
        self.freerun_handlers.clear()

        self.publishers.close()

        super(Scheduler, self).__del__()

    def _register_process_entry(self, process_entry, call_back):
        """ method parses process_entry and creates a timer_handler out of it
         timer_handler is enlisted to either :self.freerun_handlers or :self.managed_handlers
         timer_handler is started, unless it is marked as STATE_OFF """
        trigger_frequency = process_entry.trigger_frequency
        if isinstance(process_entry, ManagedProcessEntry):
            handler_key = process_entry.process_name
            handler = ManagedThreadHandler(self.logger, handler_key, trigger_frequency, call_back, process_entry)
            self.managed_handlers[handler.key] = handler
        elif isinstance(process_entry, FreerunProcessEntry):
            handler_key = (process_entry.process_name, process_entry.entry_name)
            handler = FreerunThreadHandler(self.logger, handler_key, trigger_frequency, call_back, process_entry)
            self.freerun_handlers[handler.key] = handler
        else:
            raise ValueError('ProcessEntry type {0} is not known to the system. Skipping it.'
                             .format(process_entry.__class__.__name__))

        if process_entry.is_on:
            handler.activate()
            self.logger.info('Started {0} for {1}.'
                             .format(handler.__class__.__name__, handler.key))
        else:
            self.logger.info('Registered {0} for {1}. Idle until activated.'
                             .format(handler.__class__.__name__, handler.key))

    # **************** Scheduler Methods ************************
    def _load_managed_entries(self):
        """ loads scheduler managed entries. no start-up procedures are performed """
        for process_name, process_entry in context.process_context.items():
            if isinstance(process_entry, ManagedProcessEntry):
                function = self.fire_managed_worker
            else:
                self.logger.error('Skipping non-managed context entry {0} of type {1}.'
                                  .format(process_name, process_entry.__class__.__name__))
                continue

            try:
                self._register_process_entry(process_entry, function)
            except Exception:
                self.logger.error('Managed Thread Handler {0} failed to start. Skipping it.'
                                  .format(process_entry.key), exc_info=True)

    def _load_freerun_entries(self):
        """ reads scheduler managed entries and starts their timers to trigger events """
        freerun_entries = self.freerun_process_dao.get_all()
        for freerun_entry in freerun_entries:
            try:
                self._register_process_entry(freerun_entry, self.fire_freerun_worker)
            except Exception:
                self.logger.error('Freerun Thread Handler {0} failed to start. Skipping it.'
                                  .format(freerun_entry.key), exc_info=True)

    @with_reconnect
    def start(self, *_):
        """ reads managed process entries and starts timer instances; starts dependant threads """
        db_manager.synch_db()
        self._load_managed_entries()

        try:
            self._load_freerun_entries()
        except LookupError as e:
            self.logger.warn('DB Lookup: {0}'.format(e))

        # Scheduler is initialized and running. GarbageCollector can be safely started
        self.gc.start()

        # Scheduler is initialized and running. Status Bus Listener can be safely started
        self.bus_listener.start()

        # Scheduler is initialized and running. Management Extension (MX) can be safely started
        self.mx.start()

    @thread_safe
    def fire_managed_worker(self, thread_handler_header):
        """requests next valid job for given process and manages its state"""

        def _fire_worker(process_entry, prev_job_record):
            assert isinstance(process_entry, ManagedProcessEntry)
            job_record = self.timetable.get_next_job_record(process_entry.process_name)
            state_machine = self.timetable.state_machines[process_entry.state_machine_name]
            if job_record == prev_job_record:
                # avoid the loop
                return None

            if not state_machine.run_on_active_timeperiod:
                time_qualifier = process_entry.time_qualifier
                incremented_timeperiod = time_helper.increment_timeperiod(time_qualifier, job_record.timeperiod)
                dt_record_timestamp = time_helper.synergy_to_datetime(time_qualifier, incremented_timeperiod)
                dt_record_timestamp += timedelta(minutes=LAG_5_MINUTES)

                if datetime.utcnow() <= dt_record_timestamp:
                    self.logger.info('Job {0} for {1}@{2} will not be triggered until {3}.'
                                     .format(job_record.db_id,
                                             job_record.process_name,
                                             job_record.timeperiod,
                                             dt_record_timestamp.strftime('%Y-%m-%d %H:%M:%S')))
                    return None

            blocking_type = process_entry.blocking_type
            if blocking_type == BLOCKING_DEPENDENCIES:
                state_machine.manage_job_with_blocking_dependencies(job_record)
            elif blocking_type == BLOCKING_CHILDREN:
                state_machine.manage_job_with_blocking_children(job_record)
            elif blocking_type == BLOCKING_NORMAL:
                state_machine.manage_job(job_record)
            else:
                raise ValueError('Unknown managed process type {0}'.format(blocking_type))

            return job_record

        try:
            assert isinstance(thread_handler_header, ThreadHandlerHeader)
            self.logger.info('{0} {{'.format(thread_handler_header.key))

            job_record = _fire_worker(thread_handler_header.process_entry, None)
            while job_record and job_record.is_finished:
                job_record = _fire_worker(thread_handler_header.process_entry, job_record)

        except (AMQPError, IOError) as e:
            self.logger.error('AMQPError: {0}'.format(e), exc_info=True)
            self.publishers.reset_all(suppress_logging=True)
        except Exception as e:
            self.logger.error('Exception: {0}'.format(e), exc_info=True)
        finally:
            self.logger.info('}')

    @thread_safe
    def fire_freerun_worker(self, thread_handler_header):
        """fires free-run worker with no dependencies to track"""
        try:
            assert isinstance(thread_handler_header, ThreadHandlerHeader)
            self.logger.info('{0} {{'.format(thread_handler_header.key))

            state_machine = self.timetable.state_machines[STATE_MACHINE_FREERUN]
            state_machine.manage_schedulable(thread_handler_header.process_entry)

        except Exception as e:
            self.logger.error('fire_freerun_worker: {0}'.format(e))
        finally:
            self.logger.info('}')


if __name__ == '__main__':
    from synergy.scheduler.scheduler_constants import PROCESS_SCHEDULER

    source = Scheduler(PROCESS_SCHEDULER)
    source.start()



__author__ = 'Bohdan Mushkevych'

from threading import RLock

from bson.objectid import ObjectId
from pymongo import ASCENDING
from pymongo.errors import DuplicateKeyError as MongoDuplicateKeyError

from synergy.system import time_helper
from synergy.system.time_qualifier import *
from synergy.system.decorator import thread_safe
from synergy.scheduler.scheduler_constants import COLLECTION_UNIT_OF_WORK
from synergy.conf import context
from synergy.db.error import DuplicateKeyError
from synergy.db.model import unit_of_work
from synergy.db.model.unit_of_work import UnitOfWork
from synergy.db.manager import ds_manager

QUERY_GET_FREERUN_SINCE = lambda timeperiod, include_running, include_processed, include_noop, include_failed: {
    unit_of_work.TIMEPERIOD: {'$gte': timeperiod},
    unit_of_work.UNIT_OF_WORK_TYPE: unit_of_work.TYPE_FREERUN,
    unit_of_work.STATE: {'$in': [unit_of_work.STATE_PROCESSED if include_processed else None,
                                 unit_of_work.STATE_IN_PROGRESS if include_running else None,
                                 unit_of_work.STATE_REQUESTED if include_running else None,
                                 unit_of_work.STATE_CANCELED if include_failed else None,
                                 unit_of_work.STATE_INVALID if include_failed else None,
                                 unit_of_work.STATE_NOOP if include_noop else None]}
}


class UnitOfWorkDao(object):
    """ Thread-safe Data Access Object from units_of_work table/collection """

    def __init__(self, logger):
        super(UnitOfWorkDao, self).__init__()
        self.logger = logger
        self.lock = RLock()
        self.ds = ds_manager.ds_factory(logger)

    @thread_safe
    def get_one(self, key):
        """ method finds unit_of_work record and returns it to the caller"""
        if not isinstance(key, ObjectId):
            # cast key to ObjectId
            key = ObjectId(key)

        query = {'_id': key}
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)

        document = collection.find_one(query)
        if document is None:
            msg = 'UOW with ID={0} was not found'.format(key)
            self.logger.warn(msg)
            raise LookupError(msg)
        return UnitOfWork.from_json(document)

    @thread_safe
    def get_reprocessing_candidates(self, since=None):
        """ method queries Unit Of Work whose <start_timeperiod> is younger than <since>
        and who could be candidates for re-processing """
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)

        query = {unit_of_work.STATE: {'$in': [unit_of_work.STATE_IN_PROGRESS,
                                              unit_of_work.STATE_INVALID,
                                              unit_of_work.STATE_REQUESTED]},
                 unit_of_work.UNIT_OF_WORK_TYPE: unit_of_work.TYPE_MANAGED}

        if since is None:
            cursor = collection.find(query).sort('_id', ASCENDING)
            candidates = [UnitOfWork.from_json(document) for document in cursor]
        else:
            candidates = []
            yearly_timeperiod = time_helper.cast_to_time_qualifier(QUALIFIER_YEARLY, since)
            query[unit_of_work.START_TIMEPERIOD] = {'$gte': yearly_timeperiod}

            cursor = collection.find(query).sort('_id', ASCENDING)
            for document in cursor:
                uow = UnitOfWork.from_json(document)
                if uow.process_name not in context.process_context:
                    # this is a decommissioned process
                    continue

                time_qualifier = context.process_context[uow.process_name].time_qualifier
                if time_qualifier == QUALIFIER_REAL_TIME:
                    time_qualifier = QUALIFIER_HOURLY
                process_specific_since = time_helper.cast_to_time_qualifier(time_qualifier, since)

                if process_specific_since <= uow.start_timeperiod:
                    candidates.append(uow)

        if len(candidates) == 0:
            raise LookupError('MongoDB has no UOW reprocessing candidates')
        return candidates

    @thread_safe
    def get_by_params(self, process_name, timeperiod, start_obj_id, end_obj_id):
        """ method finds unit_of_work record and returns it to the caller"""
        query = {unit_of_work.PROCESS_NAME: process_name,
                 unit_of_work.TIMEPERIOD: timeperiod,
                 unit_of_work.START_OBJ_ID: start_obj_id,
                 unit_of_work.END_OBJ_ID: end_obj_id}
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)

        document = collection.find_one(query)
        if document is None:
            raise LookupError('UOW satisfying query {0} was not found'.format(query))
        return UnitOfWork.from_json(document)

    @thread_safe
    def update(self, instance):
        """ method finds unit_of_work record and change its status"""
        assert isinstance(instance, UnitOfWork)
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)
        document = instance.document
        if instance.db_id:
            document['_id'] = ObjectId(instance.db_id)
        instance.db_id = collection.save(document, safe=True)
        return instance.db_id

    @thread_safe
    def insert(self, instance):
        """ inserts a unit of work into MongoDB.
        :raises DuplicateKeyError: if such record already exist """
        assert isinstance(instance, UnitOfWork)
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)
        try:
            return collection.insert(instance.document, safe=True)
        except MongoDuplicateKeyError as e:
            exc = DuplicateKeyError(instance.process_name,
                                    instance.start_timeperiod,
                                    instance.start_id,
                                    instance.end_id,
                                    e)
            raise exc

    @thread_safe
    def remove(self, uow_id):
        assert isinstance(uow_id, (str, ObjectId))
        collection = self.ds.connection(COLLECTION_UNIT_OF_WORK)
        return collection.remove(uow_id, safe=True)

    @thread_safe
    def run_query(self, query):
        """ method runs the query and returns a list of filtered UnitOfWork records """
        cursor = self.ds.filter(COLLECTION_UNIT_OF_WORK, query)
        return [UnitOfWork.from_json(document) for document in cursor]

    def recover_from_duplicatekeyerror(self, e):
        """ method tries to recover from DuplicateKeyError """
        if isinstance(e, DuplicateKeyError):
            try:
                return self.get_by_params(e.process_name, e.timeperiod, e.start_id, e.end_id)
            except LookupError as e:
                self.logger.error('Unable to recover from DuplicateKeyError error due to {0}'.format(e), exc_info=True)
        else:
            msg = 'Unable to recover from DuplicateKeyError due to unspecified UOW primary key'
            self.logger.error(msg)



__author__ = 'Bohdan Mushkevych'

import re
import pymongo

from synergy.db.dao.box_configuration_dao import BoxConfigurationDao, QUERY_PROCESSES_FOR_BOX_ID
from synergy.db.manager import ds_manager
from synergy.db.model.box_configuration import BoxConfiguration, BOX_ID, PROCESS_NAME
from synergy.conf import context, settings
from synergy.supervisor.supervisor_constants import PROCESS_SUPERVISOR, COLLECTION_BOX_CONFIGURATION
from synergy.system.data_logging import get_logger


CONFIG_FILE_TEMPLATE = (
    '# This file was auto-generated by Synergy Supervisor. Do not edit it manually\n'
    '# This file is expected to be located at: /etc/\n'
    '# BOX_ID identifies a box in a cluster supervised by Synergy Scheduler\n'
    'BOX_ID={0}\n'
)


def set_box_id(logger, box_id):
    config_file_name = settings.settings['config_file']
    try:
        with open(config_file_name, mode='w') as config_file:
            config_file.write(CONFIG_FILE_TEMPLATE.format(box_id))
    except Exception as e:
        logger.info('Unable to create BOX_ID file at: {0}, due to: {1}'.format(config_file_name, e))


def get_box_id(logger):
    """ retrieves box id from the /etc/synergy.conf configuration file """
    try:
        box_id = None
        config_file = settings.settings['config_file']
        with open(config_file) as a_file:
            for a_line in a_file:
                a_line = a_line.strip()
                if a_line.startswith('#'):
                    continue

                tokens = a_line.split('=')
                if tokens[0] == 'BOX_ID':
                    box_id = tokens[1]
                    return box_id

        if box_id is None:
            raise LookupError('BOX_ID is not defined in {0}'.format(config_file))

    except EnvironmentError:  # parent of IOError, OSError
        logger.error('Can not read configuration file.', exc_info=True)


class SupervisorEntry(object):
    def __init__(self, process_name):
        if process_name not in context.process_context:
            raise ValueError('SupervisorEntry: process {0} is not found in process_context. Aborting'
                             .format(process_name))

        self.logger = get_logger(PROCESS_SUPERVISOR, True)
        self.process_name = process_name
        self.re_boxes = context.process_context[process_name].present_on_boxes
        self.re_co_boxes = []  # compiled RE of boxes where the process should be present

        for re_box in self.re_boxes:
            try:
                if isinstance(re_box, str):
                    re_box = re_box.lower()
                    self.re_co_boxes.append(re.compile(re_box))
                elif isinstance(re_box, int):
                    self.re_co_boxes.append(re_box)
                else:
                    raise ValueError('SupervisorEntry support (string, integer) values. Type {0} unsupported'
                                     .format(type(re_box)))
            except TypeError:
                self.logger.warn('SupervisorEntry compilation error for {0}'.format(re_box))

    def is_present_on(self, box_id):
        box_id = box_id.lower()
        is_present = False
        for co_re_box in self.re_co_boxes:
            if isinstance(co_re_box, int):
                is_present = co_re_box == box_id

            else:
                if co_re_box.search(box_id):
                    is_present = True

            if is_present:
                break

        return is_present


class SupervisorConfigurator(object):
    def __init__(self):
        super(SupervisorConfigurator, self).__init__()
        self.logger = get_logger(PROCESS_SUPERVISOR, True)
        self.bc_dao = BoxConfigurationDao(self.logger)

        self.box_id = get_box_id(self.logger)
        self.process_map = dict()
        for process_name in context.process_context:
            try:
                self.process_map[process_name] = SupervisorEntry(process_name)
            except ValueError:
                continue

    def _change_state(self, process_name, new_state):
        self.logger.info('INFO: Supervisor configuration: setting state {0} for process {1} \n'
                         .format(new_state, process_name))
        box_config = self.bc_dao.get_one([self.box_id, process_name])
        box_config.is_on = new_state
        self.bc_dao.update(box_config)

    def mark_for_start(self, process_name):
        self._change_state(process_name, True)

    def mark_for_stop(self, process_name):
        self._change_state(process_name, False)

    def reset_db(self):
        self.logger.info('Starting *synergy.box_configuration* table reset')

        ds = ds_manager.ds_factory(self.logger)
        ds._db.drop_collection(COLLECTION_BOX_CONFIGURATION)
        self.logger.info('*synergy.box_configuration* table has been dropped')

        connection = ds.connection(COLLECTION_BOX_CONFIGURATION)
        connection.create_index([(BOX_ID, pymongo.ASCENDING), (PROCESS_NAME, pymongo.ASCENDING)], unique=True)

        for process_name, supervisor_entry in self.process_map.items():
            if supervisor_entry.is_present_on(self.box_id):
                box_config = BoxConfiguration(box_id=self.box_id,
                                              process_name=process_name,
                                              is_on=True)
                self.bc_dao.update(box_config)

        self.logger.info('*synergy.box_configuration* db has been recreated')

    def query(self):
        self.logger.info('\nSupervisor Snapshot for BOX_ID={0}:\n'.format(self.box_id))
        box_configurations = self.bc_dao.run_query(QUERY_PROCESSES_FOR_BOX_ID(self.box_id))

        for box_config in box_configurations:
            self.logger.info('{0}:\t{1}'.format(box_config.process_name, box_config.state))



"""
This module was adapted from Django's settings and configuration:
- https://raw.githubusercontent.com/django/django/master/django/conf/__init__.py
- https://raw.githubusercontent.com/django/django/master/django/utils/functional.py

Django is licensed under BSD. License is available at:
- https://github.com/django/django/blob/master/LICENSE
"""
import copy

import importlib
import os
import operator

from synergy.conf import global_context, global_settings

ENVIRONMENT_SETTINGS_VARIABLE = 'SYNERGY_SETTINGS_MODULE'
ENVIRONMENT_CONTEXT_VARIABLE = 'SYNERGY_CONTEXT_MODULE'


empty = object()


class ImproperlyConfigured(Exception):
    """Django is somehow improperly configured"""
    pass


def new_method_proxy(func):
    def inner(self, *args):
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)
    return inner


class LazyObject(object):
    """
    A wrapper for another class that can be used to delay instantiation of the
    wrapped class.

    By subclassing, you have the opportunity to intercept and alter the
    instantiation. If you don't need to do that, use SimpleLazyObject.
    """

    # Avoid infinite recursion when tracing __init__ (#19456).
    _wrapped = None

    def __init__(self):
        self._wrapped = empty

    __getattr__ = new_method_proxy(getattr)

    def __setattr__(self, name, value):
        if name == '_wrapped':
            # Assign to __dict__ to avoid infinite __setattr__ loops.
            self.__dict__['_wrapped'] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name == '_wrapped':
            raise TypeError("can't delete _wrapped.")
        if self._wrapped is empty:
            self._setup()
        delattr(self._wrapped, name)

    def _setup(self):
        """
        Must be implemented by subclasses to initialize the wrapped object.
        """
        raise NotImplementedError('subclasses of LazyObject must provide a _setup() method')

    def __deepcopy__(self, memo):
        if self._wrapped is empty:
            # We have to use type(self), not self.__class__, because the
            # latter is proxied.
            result = type(self)()
            memo[id(self)] = result
            return result
        return copy.deepcopy(self._wrapped, memo)

    __str__ = new_method_proxy(str)
    __unicode__ = new_method_proxy(unicode)
    __nonzero__ = new_method_proxy(bool)

    # Introspection support
    __dir__ = new_method_proxy(dir)

    # Need to pretend to be the wrapped class, for the sake of objects that
    # care about this (especially in equality tests)
    __class__ = property(new_method_proxy(operator.attrgetter('__class__')))
    __eq__ = new_method_proxy(operator.eq)
    __ne__ = new_method_proxy(operator.ne)
    __hash__ = new_method_proxy(hash)

    # Dictionary methods support
    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)

    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)

    def configure(self, default_settings=global_settings, **options):
        """
        Called to manually configure the settings. The 'default_settings'
        parameter sets where to retrieve any unspecified values from (its
        argument must support attribute access (__getattr__)).
        """
        if self._wrapped is not empty:
            raise RuntimeError('Settings already configured.')
        holder = UserSettingsHolder(default_settings)
        for name, value in options.items():
            setattr(holder, name, value)
        self._wrapped = holder

    @property
    def configured(self):
        """ Returns True if the settings have already been configured. """
        return self._wrapped is not empty


class LazySettings(LazyObject):
    """
    A lazy proxy for either global Synergy settings or a custom settings object.
    The user can manually configure settings prior to using them. Otherwise,
    Django uses the settings module pointed to by SYNERGY_SETTINGS_MODULE.
    """
    def _setup(self, name=None):
        """
        Load the settings module pointed to by the environment variable. This
        is used the first time we need any settings at all, if the user has not
        previously configured the settings manually.
        """
        settings_module = os.environ.get(ENVIRONMENT_SETTINGS_VARIABLE, 'settings')
        if not settings_module:
            desc = 'setting {0}'.format(name) if name else 'settings'
            raise ImproperlyConfigured(
                'Requested {0}, but settings are not configured. '
                'You must either define the environment variable {1} '
                'or call settings.configure() before accessing settings.'
                .format(desc, ENVIRONMENT_SETTINGS_VARIABLE))

        self._wrapped = Settings(settings_module, default_settings=global_settings)

    def __getattr__(self, name):
        if self._wrapped is empty:
            self._setup(name)
        return getattr(self._wrapped, name)


class LazyContext(LazyObject):
    """ A lazy proxy for Synergy Context """
    def _setup(self, name=None):
        """
        Load the settings module pointed to by the environment variable. This
        is used the first time we need any settings at all, if the user has not
        previously configured the settings manually.
        """
        settings_module = os.environ.get(ENVIRONMENT_CONTEXT_VARIABLE, 'context')
        if not settings_module:
            desc = 'setting {0}'.format(name) if name else 'settings'
            raise ImproperlyConfigured(
                'Requested {0}, but context is not configured. '
                'You must either define the environment variable {1} '
                'or call context.configure() before accessing the context.'
                .format(desc, ENVIRONMENT_CONTEXT_VARIABLE))

        self._wrapped = Settings(settings_module, default_settings=global_context)

    def __getattr__(self, name):
        if self._wrapped is empty:
            self._setup(name)
        return getattr(self._wrapped, name)


class BaseSettings(object):
    """ Common logic for settings whether set by a module or by the user. """
    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)


class Settings(BaseSettings):
    def __init__(self, settings_module, default_settings):
        # update this dict from global_settings or global_context
        for setting in dir(default_settings):
            setattr(self, setting, getattr(default_settings, setting))

        # store the settings module in case someone later cares
        self.SETTINGS_MODULE = settings_module

        mod = importlib.import_module(self.SETTINGS_MODULE)

        self._explicit_settings = set()
        for setting in dir(mod):
            setting_value = getattr(mod, setting)
            if isinstance(setting_value, dict) and hasattr(self, setting):
                existing = getattr(self, setting)
                existing.update(setting_value)
                setattr(self, setting, existing)
            else:
                setattr(self, setting, setting_value)
            self._explicit_settings.add(setting)

    def is_overridden(self, setting):
        return setting in self._explicit_settings


class UserSettingsHolder(BaseSettings):
    """ Holder for user configured settings. """
    # SETTINGS_MODULE doesn't make much sense in the manually configured (standalone) case.
    SETTINGS_MODULE = None

    def __init__(self, default_settings):
        """
        Requests for configuration variables not in this class are satisfied
        from the module specified in default_settings (if possible).
        """
        self.__dict__['_deleted'] = set()
        self.default_settings = default_settings

    def __getattr__(self, name):
        if name in self._deleted:
            raise AttributeError
        return getattr(self.default_settings, name)

    def __setattr__(self, name, value):
        self._deleted.discard(name)
        super(UserSettingsHolder, self).__setattr__(name, value)

    def __delattr__(self, name):
        self._deleted.add(name)
        if hasattr(self, name):
            super(UserSettingsHolder, self).__delattr__(name)

    def __dir__(self):
        return list(self.__dict__) + dir(self.default_settings)

    def is_overridden(self, setting):
        deleted = (setting in self._deleted)
        set_locally = (setting in self.__dict__)
        set_on_default = getattr(self.default_settings, 'is_overridden', lambda s: False)(setting)
        return deleted or set_locally or set_on_default


settings = LazySettings()
context = LazyContext()



__author__ = 'Bohdan Mushkevych'

from odm.document import BaseDocument
from odm.fields import StringField, DictField, ListField

from synergy.scheduler.scheduler_constants import EXCHANGE_UTILS


PROCESS_NAME = 'process_name'
CLASSNAME = 'classname'
MQ_QUEUE = 'mq_queue'
MQ_EXCHANGE = 'mq_exchange'
MQ_ROUTING_KEY = 'mq_routing_key'
ARGUMENTS = 'arguments'
TOKEN = 'token'
LOG_FILENAME = 'log_filename'
LOG_TAG = 'log_tag'
PID_FILENAME = 'pid_filename'
PRESENT_ON_BOXES = 'present_on_boxes'     # list of boxes where this process is monitored by the Supervisor


class DaemonProcessEntry(BaseDocument):
    """ Non-persistent model. This class presents Process Context Entry record """

    process_name = StringField(PROCESS_NAME)
    classname = StringField(CLASSNAME)
    token = StringField(TOKEN)
    mq_queue = StringField(MQ_QUEUE)
    mq_exchange = StringField(MQ_EXCHANGE)
    mq_routing_key = StringField(MQ_ROUTING_KEY)
    arguments = DictField(ARGUMENTS)
    present_on_boxes = ListField(PRESENT_ON_BOXES)
    pid_filename = StringField(PID_FILENAME)
    log_filename = StringField(LOG_FILENAME)

    @BaseDocument.key.getter
    def key(self):
        return self.process_name

    @key.setter
    def key(self, value):
        """ :param value: name of the process """
        self.process_name = value


def daemon_context_entry(process_name,
                         classname,
                         token,
                         exchange=EXCHANGE_UTILS,
                         present_on_boxes=None,
                         arguments=None,
                         queue=None,
                         routing=None,
                         pid_file=None,
                         log_file=None):
    """ forms process context entry """
    _ROUTING_PREFIX = 'routing_'
    _QUEUE_PREFIX = 'queue_'
    _SUFFIX = '_daemon'

    if queue is None:
        queue = _QUEUE_PREFIX + token + _SUFFIX
    if routing is None:
        routing = _ROUTING_PREFIX + token + _SUFFIX
    if pid_file is None:
        pid_file = token + _SUFFIX + '.pid'
    if log_file is None:
        log_file = token + _SUFFIX + '.log'
    if arguments is None:
        arguments = dict()
    else:
        assert isinstance(arguments, dict)

    process_entry = DaemonProcessEntry(
        process_name=process_name,
        classname=classname,
        token=token,
        mq_queue=queue,
        mq_routing_key=routing,
        mq_exchange=exchange,
        present_on_boxes=present_on_boxes,
        arguments=arguments,
        log_filename=log_file,
        pid_filename=pid_file)
    return process_entry
